import os
from pathlib import Path
from typing import NewType

import pytest

from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.query_scheduling.dag_collection import DagCollection
from bigquery_etl.query_scheduling.task import (
    MAX_TASK_NAME_LENGTH,
    Task,
    TaskParseException,
    TaskRef,
    UnscheduledTask,
)

TEST_DIR = Path(__file__).parent.parent


class TestTask:
    default_scheduling = {
        "dag_name": "bqetl_test_dag",
        "default_args": {"owner": "test@example.org"},
    }

    def test_of_query(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        task = Task.of_query(query_file)

        assert task.query_file == str(query_file)
        assert task.dataset == "test"
        assert task.project == "moz-fx-data-test-project"
        assert task.table == "incremental_query"
        assert task.version == "v1"
        assert task.task_name == "test__incremental_query__v1"
        assert task.dag_name == "bqetl_events"
        assert task.depends_on_past is False
        assert task.public_json
        assert task.arguments == ["--append_table"]

    def test_of_multipart_query(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "multipart_query_v1"
            / "part1.sql"
        )

        task = Task.of_multipart_query(query_file)

        assert task.query_file == str(query_file)
        assert task.dataset == "test"
        assert task.table == "multipart_query"
        assert task.project == "moz-fx-data-test-project"
        assert task.version == "v1"
        assert task.task_name == "test__multipart_query__v1"
        assert task.dag_name == "bqetl_core"
        assert task.depends_on_past is False
        assert task.multipart
        assert task.query_file_path == os.path.dirname(query_file)

    def test_of_python_script(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        task = Task.of_python_script(query_file)

        assert task.query_file == str(query_file)
        assert task.dataset == "test"
        assert task.project == "moz-fx-data-test-project"
        assert task.table == "incremental_query"
        assert task.version == "v1"
        assert task.task_name == "test__incremental_query__v1"
        assert task.dag_name == "bqetl_events"
        assert task.depends_on_past is False
        assert task.is_python_script

    def test_of_non_existing_query(self):
        with pytest.raises(FileNotFoundError):
            Task.of_query("non_existing_query/query.sql")

    def test_invalid_path_format(self):
        with pytest.raises(TaskParseException):
            metadata = Metadata("test", "test", [], {}, self.default_scheduling)
            Task.of_query(TEST_DIR / "data" / "query.sql", metadata)

    def test_unscheduled_task(self, tmp_path):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata("test", "test", [], {}, {})

        with pytest.raises(UnscheduledTask):
            Task.of_query(query_file, metadata)

    def test_no_dag_name(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata("test", "test", [], {}, self.default_scheduling)

        with pytest.raises(TaskParseException):
            Task.of_query(query_file, metadata)

    def test_task_instantiation(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_query(query_file, metadata)
        assert task.dag_name == "bqetl_test_dag"
        assert task.depends_on_past is False
        assert task.task_name == "test__incremental_query__v1"
        assert task.public_json is False
        assert task.arguments == []

    def test_task_instantiation_custom_name(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "custom_task_name",
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        task = Task.of_query(query_file, metadata)
        assert task.dag_name == "bqetl_test_dag"
        assert task.task_name == "custom_task_name"

    def test_validate_custom_task_name(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "a" * (MAX_TASK_NAME_LENGTH + 1),
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        with pytest.raises(ValueError):
            Task.of_query(query_file, metadata)

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "",
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        with pytest.raises(ValueError):
            Task.of_query(query_file, metadata)

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "task_name": "a" * MAX_TASK_NAME_LENGTH,
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        task = Task.of_query(query_file, metadata)
        assert task.task_name == "a" * MAX_TASK_NAME_LENGTH

    def test_validate_task_name(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / (("a" * MAX_TASK_NAME_LENGTH) + "_v1")
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        task = Task.of_query(query_file, metadata)
        assert task.task_name == ("a" * (MAX_TASK_NAME_LENGTH - 4)) + "__v1"

        with pytest.raises(ValueError):
            task.task_name = ("a" * MAX_TASK_NAME_LENGTH) + "__v1"
            Task.validate_task_name(task, "task_name", task.task_name)

    def test_dag_name_validation(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        with pytest.raises(ValueError):
            Task(
                dag_name="test_dag",
                query_file=str(query_file),
                owner="test@example.com",
            )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
        )

    def test_owner_validation(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        with pytest.raises(ValueError):
            assert Task(
                dag_name="bqetl_test_dag", query_file=str(query_file), owner="invalid"
            )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
        )

    def test_email_validation(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        with pytest.raises(ValueError):
            assert Task(
                dag_name="bqetl_test_dag",
                query_file=str(query_file),
                owner="test@example.com",
                email=["valid@example.com", "invalid", "valid2@example.com"],
            )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            email=[],
        )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            email=["test@example.org", "test2@example.org"],
        )

    def test_start_date_validation(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        with pytest.raises(ValueError):
            assert Task(
                dag_name="bqetl_test_dag",
                query_file=str(query_file),
                owner="test@example.com",
                start_date="March 12th 2020",
            )

        with pytest.raises(ValueError):
            assert Task(
                dag_name="bqetl_test_dag",
                query_file=str(query_file),
                owner="test@example.com",
                start_date="2020-13-12",
            )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            start_date="2020-01-01",
        )

    def test_date_partition_parameter(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            date_partition_parameter=NewType("Ignore", None),
        )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            date_partition_parameter=None,
        )

        assert Task(
            dag_name="bqetl_test_dag",
            query_file=str(query_file),
            owner="test@example.com",
            date_partition_parameter="import_date",
        )

    def test_task_get_dependencies_none(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text("SELECT 123423")

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_query(query_file, metadata)
        dags = DagCollection.from_dict({})
        task.with_upstream_dependencies(dags)
        assert task.upstream_dependencies == []
        assert task.downstream_dependencies == []

    def test_task_get_multiple_dependencies(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.table2_v1"
        )

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_query(query_file, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies

        tables = [t.task_id for t in result]

        assert "test__table1__v1" in tables
        assert "test__table2__v1" in tables

    def test_task_date_partition_offset(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.table2_v1"
        )

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        metadata_task1 = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {**self.default_scheduling, **{"date_partition_offset": -2}},
        )

        metadata_task2 = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {**self.default_scheduling, **{"date_partition_offset": -1}},
        )

        task = Task.of_query(query_file, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        table1_task = Task.of_query(table1_file, metadata_task1)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata_task2)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies

        assert task.date_partition_offset == -2
        assert [
            t.date_partition_offset for t in result if t.task_id == "test__table1__v1"
        ] == [-2]
        assert [
            t.date_partition_offset for t in result if t.task_id == "test__table2__v1"
        ] == [-1]
        assert task.date_partition_parameter == "submission_date"

    def test_task_date_partition_offset_recursive(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text("SELECT * FROM `test-project`.test.table1_v1")

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        table2_metadata = Metadata(
            "test",
            "test",
            ["test@example.org"],
            {},
            {**self.default_scheduling, **{"date_partition_offset": -2}},
        )

        task = Task.of_query(query_file, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT * FROM `test-project`.test.table2_v1")
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, table2_metadata)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies

        assert task.date_partition_offset == -2
        assert [
            t.date_partition_offset for t in result if t.task_id == "test__table1__v1"
        ] == [-2]
        assert task.date_partition_parameter == "submission_date"

    def test_multipart_task_get_dependencies(self, tmp_path):
        query_dir = tmp_path / "test-project" / "test" / "query_v1"
        os.makedirs(query_dir)

        query_file_part1 = query_dir / "part1.sql"
        query_file_part1.write_text("SELECT * FROM `test-project`.test.table1_v1")

        query_file_part2 = query_dir / "part2.sql"
        query_file_part2.write_text("SELECT * FROM `test-project`.test.table2_v1")

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_multipart_query(query_file_part1, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies

        tables = [t.task_id for t in result]

        assert "test__table1__v1" in tables
        assert "test__table2__v1" in tables

    def test_task_get_view_dependencies(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.test_view"
        )

        view_file = tmp_path / "test-project" / "test" / "test_view" / "view.sql"
        os.makedirs(view_file.parent)
        view_file.write_text(
            "CREATE OR REPLACE VIEW `test-project`.test.test_view "
            "AS SELECT * FROM `test-project`.test.table2_v1"
        )

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_query(query_file, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies

        tables = [t.task_id for t in result]

        assert "test__table1__v1" in tables
        assert "test__table2__v1" in tables

    def test_task_get_nested_view_dependencies(self, tmp_path):
        query_file = tmp_path / "test-project" / "test" / "query_v1" / "query.sql"
        os.makedirs(query_file.parent)
        query_file.write_text(
            "SELECT * FROM `test-project`.test.table1_v1 "
            "UNION ALL SELECT * FROM `test-project`.test.test_view"
        )

        view_file = tmp_path / "test-project" / "test" / "test_view" / "view.sql"
        os.makedirs(view_file.parent)
        view_file.write_text(
            "CREATE OR REPLACE VIEW `test-project`.test.test_view "
            "AS SELECT * FROM `test-project`.test.test_view2"
        )

        view2_file = tmp_path / "test-project" / "test" / "test_view2" / "view.sql"
        os.makedirs(view2_file.parent)
        view2_file.write_text(
            "CREATE OR REPLACE VIEW `test-project`.test.test_view2 "
            "AS SELECT * FROM `test-project`.test.table2_v1"
        )

        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        task = Task.of_query(query_file, metadata)

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata)

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([task, table1_task, table2_task])

        task.with_upstream_dependencies(dags)
        result = task.upstream_dependencies
        tables = [t.task_id for t in result]

        assert "test__table1__v1" in tables
        assert "test__table2__v1" in tables

    def test_task_depends_on(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        scheduling = {
            "dag_name": "bqetl_test_dag",
            "default_args": {"owner": "test@example.org"},
            "depends_on": [
                {"dag_name": "external_dag", "task_id": "external_task"},
                {
                    "dag_name": "external_dag2",
                    "task_id": "external_task2",
                    "execution_delta": "15m",
                },
            ],
        }

        metadata = Metadata("test", "test", ["test@example.org"], {}, scheduling)

        task = Task.of_query(query_file, metadata)
        assert task.dag_name == "bqetl_test_dag"
        assert len(task.depends_on) == 2
        assert task.depends_on[0].dag_name == "external_dag"
        assert task.depends_on[0].task_id == "external_task"

        assert task.depends_on[1].dag_name == "external_dag2"
        assert task.depends_on[1].task_id == "external_task2"
        assert task.depends_on[1].execution_delta == "15m"

    def test_task_trigger_rule(self):
        query_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "query.sql"
        )

        task = Task(
            dag_name="bqetl_test_dag",
            owner="test@example.org",
            query_file=str(query_file),
            trigger_rule="all_success",
        )

        assert task.trigger_rule == "all_success"

        with pytest.raises(ValueError, match=r"Invalid trigger rule an_invalid_rule"):
            assert Task(
                dag_name="bqetl_test_dag",
                owner="test@example.org",
                query_file=str(query_file),
                trigger_rule="an_invalid_rule",
            )

    def test_task_ref(self):
        task_ref = TaskRef(dag_name="test_dag", task_id="task")

        assert task_ref.dag_name == "test_dag"
        assert task_ref.task_id == "task"
        assert task_ref.execution_delta is None

        task_ref = TaskRef(dag_name="test_dag", task_id="task", execution_delta="15m")

        assert task_ref.execution_delta == "15m"

    def test_task_ref_execution_delta_validation(self):
        with pytest.raises(ValueError):
            TaskRef(dag_name="test_dag", task_id="task", execution_delta="invalid")

        assert TaskRef(dag_name="test_dag", task_id="task", execution_delta="1h15m")

    def test_check_get_dependencies(self, tmp_path):
        metadata = Metadata(
            "test", "test", ["test@example.org"], {}, self.default_scheduling
        )

        table1_file = tmp_path / "test-project" / "test" / "table1_v1" / "query.sql"
        os.makedirs(table1_file.parent)
        table1_file.write_text("SELECT 12345")
        metadata.write(
            tmp_path / "test-project" / "test" / "table1_v1" / "metadata.yaml"
        )
        table1_task = Task.of_query(table1_file, metadata)

        table2_file = tmp_path / "test-project" / "test" / "table2_v1" / "query.sql"
        os.makedirs(table2_file.parent)
        table2_file.write_text("SELECT 67890")
        table2_task = Task.of_query(table2_file, metadata)

        check_file = tmp_path / "test-project" / "test" / "table1_v1" / "checks.sql"
        check_file.write_text("SELECT * FROM `test-project`.test.table2_v1")
        checks_task = Task.of_dq_check(
            check_file, is_check_fail=True, metadata=metadata
        )

        dags = DagCollection.from_dict(
            {
                "bqetl_test_dag": {
                    "schedule_interval": "daily",
                    "default_args": {
                        "owner": "test@example.org",
                        "start_date": "2020-01-01",
                    },
                }
            }
        ).with_tasks([checks_task, table1_task, table2_task])

        checks_task.with_upstream_dependencies(dags)
        result = checks_task.upstream_dependencies

        tables = [t.task_id for t in result]

        assert len(tables) == 2
        assert "test__table1__v1" in tables
        assert "test__table2__v1" in tables



from datetime import date

from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.validate_metadata import (
    validate_deprecation,
    validate_public_data,
)


class TestValidateMetadata(object):
    def test_is_valid_public_data(self):
        metadata_not_public = Metadata("No public data", "No public data", {}, {})
        assert validate_public_data(metadata_not_public, "test/path/metadata.yaml")

        metadata_valid_public = Metadata(
            "Public json data",
            "Public json data",
            [],
            {"public_json": True, "review_bugs": [123456]},
            {},
        )
        assert validate_public_data(metadata_valid_public, "test/path/metadata.yaml")

        metadata_valid_public = Metadata(
            "Public BigQuery data",
            "Public BigQuery data",
            [],
            {"public_bigquery": True, "review_bugs": [123456]},
            {},
        )
        assert validate_public_data(metadata_valid_public, "test/path/metadata.yaml")

        metadata_invalid_public = Metadata(
            "Public BigQuery data",
            "Public BigQuery data",
            [],
            {"public_bigquery": True},
            {},
        )
        assert (
            validate_public_data(metadata_invalid_public, "test/path/metadata.yaml")
            is False
        )

    def test_validate_deprecation(self):
        metadata_valid = Metadata(
            friendly_name="test",
            description="test",
            owners=["test@example.org"],
            labels={"test": "true", "foo": "abc"},
            deprecated=True,
            deletion_date=date(2024, 5, 4),
        )

        assert validate_deprecation(metadata_valid, "test/path/metadata.yaml")

        metadata_valid = Metadata(
            friendly_name="test",
            description="test",
            owners=["test@example.org"],
            labels={"test": "true", "foo": "abc"},
            deprecated=True,
            deletion_date=None,
        )

        assert validate_deprecation(metadata_valid, "test/path/metadata.yaml")

        metadata_valid = Metadata(
            friendly_name="test",
            description="test",
            owners=["test@example.org"],
            labels={"test": "true", "foo": "abc"},
            deprecated=False,
            deletion_date=date(2024, 5, 4),
        )

        assert not validate_deprecation(metadata_valid, "test/path/metadata.yaml")



"""Testing data for query."""

from itertools import product
from pathlib import Path

import yaml

ROOT = Path(__file__).parent

UUID = "df735f02-efe5-4b07-b212-583bb99ba241"
SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"
OS = "Android"
PING_TYPE = "metrics"

# See the scalar_bucket_counts minimal example for more details on the
# preconditions.
CLIENTS_HISTOGRAM_AGGREGATES = [
    {
        "sample_id": 1,
        "client_id": UUID,
        "ping_type": PING_TYPE,
        "os": OS,
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "*",
        "histogram_aggregates": [
            {
                "metric": "network_tcp_connection",
                "metric_type": "timing_distribution",
                "key": "",
                "agg_type": "summed_histogram",
                "value": [
                    {"key": "1", "value": 1},
                    {"key": "2", "value": 0},
                ],
            }
        ],
    }
]

BASE_ROW = {
    "agg_type": "summed_histogram",
    "app_build_id": "*",
    "app_version": 84,
    "channel": "*",
    "key": "",
    "metric": "network_tcp_connection",
    "metric_type": "timing_distribution",
    "os": "*",
    "ping_type": "*",
    "range_max": 2,
    "record": {"key": "1", "value": 1.0},
}

EXPECT = [
    {**BASE_ROW, **dict(zip(["record", "ping_type", "os", "app_build_id"], values))}
    for values in product(
        [{"key": "1", "value": 1.0}, {"key": "2", "value": 0.0}],
        *zip([PING_TYPE, OS, APP_BUILD_ID], ["*"] * 3),
    )
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__clients_histogram_aggregates_v1.yaml",
        CLIENTS_HISTOGRAM_AGGREGATES,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""Upload changes in Suppression list to Campaign Monitor API."""

from pathlib import Path
from typing import List

import requests
import rich_click as click
from google.cloud import bigquery
from requests import HTTPError
from requests.auth import HTTPBasicAuth

from bigquery_etl.schema import Schema

BASE_URL = "https://api.createsend.com/api/v3.3/clients"
UPLOAD_BATCH_SIZE = 100

PROJECT_ID = Path(__file__).parent.parent.parent.name
DATASET_ID = Path(__file__).parent.parent.name
TABLE_NAME = Path(__file__).parent.name

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = Schema.from_schema_file(SCHEMA_FILE).to_bigquery_schema()


def list_new_suppressions() -> List[str]:
    """Download the new_suppression_list_entries_for_mofo_v1 table into an array."""
    client = bigquery.Client()
    query = """
        SELECT
            main.email
        FROM
            `moz-fx-data-shared-prod.marketing_suppression_list_derived.main_suppression_list_v1` AS main
        LEFT JOIN
            `moz-fx-data-shared-prod.marketing_suppression_list_external.campaign_monitor_suppression_list_v1` AS current_mofo
            ON main.email = current_mofo.email
        LEFT JOIN
            `moz-fx-data-shared-prod.marketing_suppression_list_external.send_suppression_list_update_to_campaign_monitor_v1` AS not_uploadable
            ON main.email = not_uploadable.email
        WHERE
            current_mofo.email IS NULL
            AND not_uploadable.email IS NULL
    """
    data = client.query(query).result()
    return [row["email"] for row in list(data)]


def upload_new_suppressions_to_campaign_monitor(
    base_url: str, client_id: str, api_key: str, emails: List[str]
) -> None:
    """Add emails to suppression list in Campaign Monitor."""
    auth = HTTPBasicAuth(api_key, "")
    add_suppression_list_url = f"{base_url}/{client_id}/suppress.json"
    payload = {"EmailAddresses": emails}
    response = requests.post(url=add_suppression_list_url, auth=auth, json=payload)
    response.raise_for_status()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    click.echo(f"Loaded {len(data)} new rows into {destination_table_id}.")


@click.command
@click.option(
    "--api_key",
    required=True,
    help="Campaign Monitor API key to use for authentication.",
)
@click.option(
    "--client_id",
    required=True,
    help="Client Id for Campaign Monitor.",
)
def main(api_key: str, client_id: str) -> None:
    """Download list of new suppression from BigQuery and upload it to the Campaign Monitor API."""
    base_url = BASE_URL
    # Get emails to upload
    emails = list_new_suppressions()
    number_of_emails = len(emails)

    # try to upload in UPLOAD_BATCH_SIZE chunks
    not_uploaded_emails = []
    uploaded = 0
    for i in range(0, number_of_emails, UPLOAD_BATCH_SIZE):
        partial_emails = emails[i : min(i + UPLOAD_BATCH_SIZE, number_of_emails)]
        try:
            upload_new_suppressions_to_campaign_monitor(
                base_url=base_url,
                api_key=api_key,
                client_id=client_id,
                emails=partial_emails,
            )
            uploaded += len(partial_emails)
            click.echo(f"Uploaded {uploaded}/{number_of_emails} emails.")
        except HTTPError as err:
            click.echo(err)
            not_uploaded_emails += partial_emails

    # try to upload the emails that were not able to upload in batch one by one
    not_uploadable_emails = []
    for email in not_uploaded_emails:
        try:
            upload_new_suppressions_to_campaign_monitor(
                base_url=base_url, api_key=api_key, client_id=client_id, emails=[email]
            )
        except HTTPError as err:
            click.echo(err)
            not_uploadable_emails.append({"email": email})

    # add not uploadable emails in table
    if len(not_uploadable_emails) > 0:
        destination_table_id = f"{PROJECT_ID}.{DATASET_ID}.{TABLE_NAME}"
        store_data_in_bigquery(
            data=not_uploadable_emails,
            schema=SCHEMA,
            destination_project=PROJECT_ID,
            destination_table_id=destination_table_id,
        )
        click.echo(
            f"Added {len(not_uploadable_emails)} emails to the list of emails that can't be uploaded due to bad formatting."
        )


if __name__ == "__main__":
    main()



from pathlib import Path

import bigquery_etl.query_scheduling.generate_airflow_dags as gad

TEST_DIR = Path(__file__).parent.parent


class TestGenerateAirflowDags(object):
    sql_dir = TEST_DIR / "data" / "test_sql"
    dags_config = TEST_DIR / "data" / "dags.yaml"

    def test_get_dags(self):
        dags = gad.get_dags(self.sql_dir, self.dags_config)

        assert len(dags.dags) == 2
        assert dags.dag_by_name("not existing") is None
        assert dags.dag_by_name("bqetl_events") is not None
        assert dags.dag_by_name("bqetl_core") is not None

        events_dag = dags.dag_by_name("bqetl_events")
        assert len(events_dag.tasks) == 3
        assert events_dag.tasks[0].depends_on_past is False
        assert events_dag.tasks[1].depends_on_past is False

        core_dag = dags.dag_by_name("bqetl_core")
        assert len(core_dag.tasks) == 4



"""Adjust data - download deliverables, clean and upload to BigQuery."""

import csv
import json
import tempfile
from argparse import ArgumentParser

import requests
from google.cloud import bigquery

API_URI = "api.adjust.com"
ENDPOINT = "kpis"
API_VERSION = "v1"

"""The APP_TOKEN_LIST is a list of dictionaries.
Keys are the app_name, and app_token in the form:
{"app_name":"<specific app name from Adjust dash>". "app_token":"<unique Adjust app token>}"
Look here to see the apps we are tracking: https://dash.adjust.com/#/.
It is important to maintain this list in order for the script to work especially in the case of new apps
being added to track in Adjust
"""

CSV_FIELDS = [
    "cohort_start_month",
    "period_length",
    "period",
    "app",
    "app_token",
    "network",
    "network_token",
    "country",
    "os",
    "cohort_size",
    "retained_users",
    "retention_rate",
    "time_spent_per_user",
    "time_spent_per_session",
    "time_spent",
    "sessions_per_user",
    "sessions",
    "date",
]


def read_json(filename: str) -> dict:
    """Read JSON file."""
    with open(filename, "r") as f:
        data = json.loads(f.read())
    return data


def write_dict_to_csv(json_data, filename):
    """Write a dictionary to a csv."""
    with open(filename, "w") as out_file:
        dict_writer = csv.DictWriter(out_file, CSV_FIELDS)
        dict_writer.writeheader()
        dict_writer.writerows(json_data)


def download_adjust_kpi_data(date, api_token, app_token):
    """Download data from Adjust - API token and APP tokens are called here."""
    start_date = "2020-09-01"  # use date mozilla migrated fennec to fenix as start date
    end_date = date

    kpis = [
        "cohort_size",
        "retained_users",
        "retention_rate",
        "time_spent_per_user",
        "time_spent_per_session",
        "time_spent",
        "sessions_per_user",
        "sessions",
    ]
    groupings = [
        "months",
        "periods",  # from result_parameter
        "apps",
        "networks",
        "countries",
        "os_names",
    ]
    # getting overview metrics for different kpis / Deliverables
    url = f"https://api.adjust.com/kpis/v1/{app_token}/cohorts"  # overview
    url_params = f"start_date={start_date}&end_date={end_date}&kpis={','.join(kpis)}&grouping={','.join(groupings)}"
    headers = {
        "Authorization": f"Bearer {api_token}",
    }

    response = requests.get(url, headers=headers, params=url_params)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)

    return response


def check_json(adjust_response_text):
    """Script will return an empty dictionary for apps on days when there is no data. Check for that here."""
    with tempfile.NamedTemporaryFile() as tmp_json:
        with open(tmp_json.name, "w") as f_json:
            f_json.write(adjust_response_text)
            try:
                query_export = read_json(f_json.name)
            except (
                ValueError
            ):  # ex. json.decoder.JSONDecodeError: Expecting value: line 1 column 1 (char 0)
                return None
    return query_export


def clean_json(query_export, date_script_run):
    """JSON sometimes has missing keys, need to clean up the data. Cohort_start_date should always be the first of the month. Date or end_date is day script is run.

    https://help.adjust.com/en/article/kpi-glossary#cohort-metrics has descriptions of groupings and metrics.
    """
    period_length = query_export["result_parameters"]["period"]
    run_date = date_script_run

    fields_list = []
    for date in query_export["result_set"]["dates"]:
        r_date = date["date"]
        for period in date["periods"]:
            r_period_length = period_length
            r_period = period.get("period", "no_periods")
            for app in period["apps"]:
                r_app = app.get("name", "no_app_name")
                r_app_token = app.get("token", "no_app_token")
                for network in app["networks"]:
                    r_network = network.get("name", "no_network_name")
                    r_network_token = network.get("token", "no_network_token")
                    for country in network["countries"]:
                        r_country = country["country"]
                        for os_name in country["os_names"]:
                            r_os_name = os_name["os_name"]
                            field_dict = {
                                "cohort_start_month": (r_date),
                                "period_length": (r_period_length),
                                "period": (r_period),
                                "app": (r_app),
                                "app_token": (r_app_token),
                                "network": (r_network),
                                "network_token": (r_network_token),
                                "country": (r_country),
                                "os": (r_os_name),
                                "cohort_size": os_name["kpi_values"][0],
                                "retained_users": os_name["kpi_values"][1],
                                "retention_rate": os_name["kpi_values"][2],
                                "time_spent_per_user": os_name["kpi_values"][3],
                                "time_spent_per_session": os_name["kpi_values"][4],
                                "time_spent": os_name["kpi_values"][5],
                                "sessions_per_user": os_name["kpi_values"][6],
                                "sessions": os_name["kpi_values"][7],
                                "date": run_date,
                            }
                            fields_list.append(field_dict)

    return fields_list


def upload_to_bigquery(csv_data, project, dataset, date):
    """Upload the data to bigquery."""
    print("writing json to csv")

    partition = f"{date}".replace("-", "")

    with tempfile.NamedTemporaryFile() as tmp_csv:
        with open(tmp_csv.name, "w+b") as f_csv:
            write_dict_to_csv(csv_data, f_csv.name)

            client = bigquery.Client(project)

            job_config = bigquery.LoadJobConfig(
                create_disposition="CREATE_IF_NEEDED",
                write_disposition="WRITE_TRUNCATE",
                time_partitioning=bigquery.TimePartitioning(
                    type_=bigquery.TimePartitioningType.DAY,
                    field="date",
                ),
                skip_leading_rows=1,
                schema=[
                    bigquery.SchemaField("cohort_start_month", "DATE"),
                    bigquery.SchemaField("period_length", "STRING"),
                    bigquery.SchemaField("period", "INTEGER"),
                    bigquery.SchemaField("app", "STRING"),
                    bigquery.SchemaField("app_token", "STRING"),
                    bigquery.SchemaField("network", "STRING"),
                    bigquery.SchemaField("network_token", "STRING"),
                    bigquery.SchemaField("country", "STRING"),
                    bigquery.SchemaField("os", "STRING"),
                    bigquery.SchemaField("cohort_size", "INTEGER"),
                    bigquery.SchemaField("retained_users", "INTEGER"),
                    bigquery.SchemaField("retention_rate", "FLOAT"),
                    bigquery.SchemaField("time_spent_per_user", "INTEGER"),
                    bigquery.SchemaField("time_spent_per_session", "INTEGER"),
                    bigquery.SchemaField("time_spent", "INTEGER"),
                    bigquery.SchemaField("sessions_per_user", "FLOAT"),
                    bigquery.SchemaField("sessions", "INTEGER"),
                    bigquery.SchemaField("date", "DATE"),
                ],
            )
            # Table names are based on the app name seen in the Adjust dashboard"
            destination = f"{project}.{dataset}.adjust_cohort_v1${partition}"

            job = client.load_table_from_file(f_csv, destination, job_config=job_config)

            print(
                f"Writing adjust data for all apps to {destination}. BigQuery job ID: {job.job_id}"
            )
            job.result()


def main():
    """Input data, call functions, get stuff done."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--adjust_api_token", required=True)
    parser.add_argument("--adjust_app_list", required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="adjust_derived")

    args = parser.parse_args()

    app_list = json.loads(args.adjust_app_list)

    data = []

    # Cycle through the apps to get the relevant kpi data
    for app in app_list:
        print(f'This is data for {app["app_name"]}')
        # Ping the Adjust URL and get a response
        json_file = download_adjust_kpi_data(
            args.date,
            args.adjust_api_token,
            app["app_token"],
        )

        query_export = check_json(json_file.text)

        if query_export is not None:
            # This section writes the tmp json data into a temp CSV file which will then be put into a BigQuery table
            adjust_data = clean_json(query_export, args.date)
            data.extend(adjust_data)
        else:
            print(f'no data for {app["app_name"]} today')

    upload_to_bigquery(data, args.project, args.dataset, args.date)


if __name__ == "__main__":
    main()



import os

import pytest

from bigquery_etl.dryrun import DryRun, Errors


@pytest.fixture
def tmp_query_path(tmp_path):
    p = tmp_path / "moz-fx-data-shared-prod" / "telemetry_derived" / "mytable"
    p.mkdir(parents=True)
    return p


class TestDryRun:
    def test_dry_run_sql_file(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        response = dryrun.dry_run_result
        assert response["valid"]

    def test_dry_run_invalid_sql_file(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text("SELECT INVALID 123")

        dryrun = DryRun(str(query_file))
        response = dryrun.dry_run_result
        assert response["valid"] is False

    def test_sql_file_valid(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.is_valid()

    def test_view_file_valid(self, tmp_query_path):
        view_file = tmp_query_path / "view.sql"
        view_file.write_text(
            """
            SELECT
            *
            FROM
            `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
        """
        )

        # this view file is only valid with strip_dml flag
        dryrun = DryRun(sqlfile=str(view_file), strip_dml=True)
        assert dryrun.get_error() is Errors.DATE_FILTER_NEEDED
        assert dryrun.is_valid()

    def test_sql_file_invalid(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text("SELECT INVALID 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.is_valid() is False

    def test_get_referenced_tables_empty(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text("SELECT 123")

        dryrun = DryRun(str(query_file))
        assert dryrun.get_referenced_tables() == []

    def test_get_sql(self, tmp_path):
        os.makedirs(tmp_path / "telmetry_derived")
        query_file = tmp_path / "telmetry_derived" / "query.sql"

        sql_content = "SELECT 123 "
        query_file.write_text(sql_content)

        assert DryRun(sqlfile=str(query_file)).get_sql() == sql_content
        with pytest.raises(ValueError):
            DryRun(sqlfile="invalid path").get_sql()

    def test_get_referenced_tables(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text(
            "SELECT * FROM telemetry_derived.clients_daily_v6 "
            "WHERE submission_date = '2020-01-01'"
        )
        query_dryrun = DryRun(str(query_file)).get_referenced_tables()

        assert len(query_dryrun) == 1
        assert query_dryrun[0]["datasetId"] == "telemetry_derived"
        assert query_dryrun[0]["tableId"] == "clients_daily_v6"

        view_file = tmp_query_path / "view.sql"
        view_file.write_text(
            """
            CREATE OR REPLACE VIEW
            `moz-fx-data-shared-prod.telemetry.clients_daily`
            AS
            SELECT
            *
            FROM
            `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
        """
        )
        view_dryrun = DryRun(str(view_file), strip_dml=True).get_referenced_tables()

        assert len(view_dryrun) == 1
        assert view_dryrun[0]["datasetId"] == "telemetry_derived"
        assert view_dryrun[0]["tableId"] == "clients_daily_v6"

        view_file.write_text(
            """
        SELECT document_id
        FROM mozdata.org_mozilla_firefox.baseline
        WHERE submission_timestamp > current_timestamp()
        UNION ALL
        SELECT document_id
        FROM mozdata.org_mozilla_fenix.baseline
        WHERE submission_timestamp > current_timestamp()
        """
        )
        multiple_tables = DryRun(str(view_file)).get_referenced_tables()
        multiple_tables.sort(key=lambda x: x["datasetId"])

        assert len(multiple_tables) == 2
        assert multiple_tables[0]["datasetId"] == "org_mozilla_fenix_stable"
        assert multiple_tables[0]["tableId"] == "baseline_v1"
        assert multiple_tables[1]["datasetId"] == "org_mozilla_firefox_stable"
        assert multiple_tables[1]["tableId"] == "baseline_v1"

    def test_get_error(self, tmp_query_path):
        view_file = tmp_query_path / "view.sql"

        view_file.write_text(
            """
        CREATE OR REPLACE VIEW
          `moz-fx-data-shared-prod.telemetry.clients_daily`
        AS
        SELECT
        *
        FROM
          `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
        """
        )

        valid_dml_stripped = """
        SELECT
        *
        FROM
          `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
        WHERE submission_date > current_date()
        """

        invalid_dml_stripped = """
        SELECT
        *
        FROM
          `moz-fx-data-shared-prod.telemetry_derived.clients_daily_v6`
        WHERE something
        WHERE submission_date > current_date()
        """

        assert DryRun(sqlfile=str(view_file)).get_error() is Errors.READ_ONLY
        assert (
            DryRun(sqlfile=str(view_file), strip_dml=True).get_error()
            is Errors.DATE_FILTER_NEEDED
        )
        assert (
            DryRun(sqlfile=str(view_file), content=invalid_dml_stripped).get_error()
            is Errors.DATE_FILTER_NEEDED_AND_SYNTAX
        )
        assert (
            DryRun(
                sqlfile=str(view_file), content=valid_dml_stripped, strip_dml=True
            ).get_error()
            is None
        )

    def test_dryrun_metrics_query(self, tmp_query_path):
        query_file = tmp_query_path / "query.sql"
        query_file.write_text(
            """
            SELECT * FROM (
                {{ metrics.calculate(
                    metrics=['days_of_use', 'uri_count', 'ad_clicks'],
                    platform='firefox_desktop',
                    group_by={'sample_id': 'sample_id'},
                    where='submission_date = "2023-01-01"'
                ) }}
            )
            """
        )

        dryrun = DryRun(sqlfile=str(query_file))
        assert dryrun.is_valid()



import json
from pathlib import Path
from unittest.mock import MagicMock, Mock, call

import pytest
import smart_open

from bigquery_etl.public_data.publish_json import JsonPublisher

TEST_DIR = Path(__file__).parent.parent


class TestPublishJson(object):
    test_bucket = "test-bucket"
    project_id = "test-project-id"

    non_incremental_sql_path = (
        TEST_DIR
        / "data"
        / "test_sql"
        / "moz-fx-data-test-project"
        / "test"
        / "non_incremental_query_v1"
        / "query.sql"
    )

    incremental_sql_path = (
        TEST_DIR
        / "data"
        / "test_sql"
        / "moz-fx-data-test-project"
        / "test"
        / "incremental_query_v1"
        / "query.sql"
    )

    incremental_parameter = "submission_date:DATE:2020-03-15"

    mock_blob = Mock()
    mock_blob.download_as_string.return_value = bytes("[]", "utf-8")
    mock_blob.name = "blob_path/000000000000.ndjson"

    mock_bucket = Mock()
    mock_bucket.blob.return_value = mock_blob
    mock_bucket.rename_blob.return_value = None

    extract_table_mock = Mock()
    extract_table_mock.result.return_value = None

    mock_client = Mock()
    mock_client.get_table.return_value = None
    mock_client.extract_table.return_value = extract_table_mock

    mock_storage_client = Mock()
    mock_storage_client.list_blobs.return_value = [mock_blob]
    mock_storage_client.bucket.return_value = mock_bucket

    temp_table = f"{project_id}.tmp.incremental_query_v1_20200315"
    non_incremental_table = f"{project_id}.test.non_incremental_query_v1"
    api_version = "v1"

    @pytest.fixture(autouse=True)
    def setup(self):
        pass

    def test_dataset_table_version_splitting(self):
        publisher = JsonPublisher(
            self.mock_client,
            self.mock_storage_client,
            self.project_id,
            str(self.non_incremental_sql_path),
            self.api_version,
            self.test_bucket,
        )
        assert publisher.dataset == "test"
        assert publisher.table == "non_incremental_query"
        assert publisher.version == "v1"

    def test_invalid_query_file(self):
        with pytest.raises(FileNotFoundError) as e:
            JsonPublisher(
                self.mock_client,
                self.mock_storage_client,
                self.project_id,
                "invalid_path",
                self.api_version,
                self.test_bucket,
            )

        assert e.type == FileNotFoundError

    def test_parameter_date(self):
        publisher = JsonPublisher(
            self.mock_client,
            self.mock_storage_client,
            self.project_id,
            str(self.incremental_sql_path),
            self.api_version,
            self.test_bucket,
            ["submission_date:DATE:2020-11-02"],
        )

        assert publisher.date == "2020-11-02"

    def test_no_date_parameter_incremental(self):
        with pytest.raises(SystemExit) as e:
            publisher = JsonPublisher(
                self.mock_client,
                self.mock_storage_client,
                self.project_id,
                str(self.incremental_sql_path),
                self.api_version,
                self.test_bucket,
                None,
            )
            publisher.publish_json()

        assert e.type == SystemExit
        assert e.value.code == 1

    def test_write_results_to_temp_table(self):
        publisher = JsonPublisher(
            self.mock_client,
            self.mock_storage_client,
            self.project_id,
            str(self.incremental_sql_path),
            self.api_version,
            self.test_bucket,
            ["submission_date:DATE:2020-03-15"],
        )
        publisher._write_results_to_temp_table()

        assert publisher.temp_table.startswith(self.temp_table)
        self.mock_client.query.assert_called_once()

    def test_gcp_convert_ndjson_to_json(self):
        publisher = JsonPublisher(
            self.mock_client,
            self.mock_storage_client,
            self.project_id,
            str(self.incremental_sql_path),
            self.api_version,
            self.test_bucket,
            ["submission_date:DATE:2020-03-15"],
        )

        mock_out = MagicMock(side_effect=[['{"a": 1}', '{"b": "cc"}'], None])
        mock_out.__iter__ = Mock(return_value=iter(['{"a": 1}', '{"b": "cc"}']))
        file_handler = MagicMock()
        file_handler.__enter__.return_value = mock_out

        smart_open.open = MagicMock(return_value=file_handler)

        publisher._gcp_convert_ndjson_to_json("test_path")

        smart_open.open.assert_has_calls(
            [
                call("gs://test-bucket/blob_path/000000000000.ndjson"),
                call("gs://test-bucket/blob_path/000000000000.json.tmp.gz", "w"),
            ]
        )

        file_handler.write.assert_has_calls(
            [call("["), call('{"a": 1}'), call(","), call('{"b": "cc"}'), call("]")]
        )

    def test_publish_last_updated_to_gcs(self):
        publisher = JsonPublisher(
            self.mock_client,
            self.mock_storage_client,
            self.project_id,
            str(self.incremental_sql_path),
            self.api_version,
            self.test_bucket,
            ["submission_date:DATE:2020-03-15"],
        )

        mock_out = MagicMock(side_effect=[['{"a": 1}', '{"b": "cc"}'], None])
        mock_out.__iter__ = Mock(return_value=iter(['{"a": 1}', '{"b": "cc"}']))
        file_handler = MagicMock()
        file_handler.__enter__.return_value = mock_out

        smart_open.open = MagicMock(return_value=file_handler)

        publisher.publish_json()

        assert publisher.last_updated is not None
        mock_out.write.assert_called_with(
            json.dumps(publisher.last_updated.strftime("%Y-%m-%d %H:%M:%S"))
        )



# Import release data from https://product-details.mozilla.org/1.0/mobile_android.json

from argparse import ArgumentParser
from google.cloud import bigquery
import io
import json
import requests
from pathlib import Path
import yaml

URL = "https://product-details.mozilla.org/1.0/mobile_android.json"
SCHEMA_FILE = Path(__file__).parent / "schema.yaml"

parser = ArgumentParser(description=__doc__)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="org_mozilla_fenix_derived")
parser.add_argument("--destination_table", default="releases_v1")


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    release_json = json.loads(requests.get(URL).text)
    # ignoring the key here since it is <product>-<version>
    release_data = [r for _, r in release_json["releases"].items()]

    schema_dict = yaml.safe_load(SCHEMA_FILE.read_text())["fields"]
    schema = client.schema_from_json(io.StringIO(json.dumps(schema_dict)))

    job_config = bigquery.LoadJobConfig(schema=schema)
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE

    job = client.load_table_from_json(
        release_data,
        f"{args.destination_dataset}.{args.destination_table}",
        job_config=job_config,
    )
    job.result()


if __name__ == "__main__":
    main()



"""PyTest plugins for running linters on python scripts without a .py extension."""

import re

python_hashbang = re.compile(r"#!(/usr)?/bin(/env |/)python(3(.7)?)?$")


def is_python_executable(path):
    """Check if the given path is a python script."""
    if path.ext == "":
        with open(path) as fp:
            return python_hashbang.match(fp.readline().strip())
    return False



"""Testing data for query."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent

UUID = "df735f02-efe5-4b07-b212-583bb99ba241"
SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

VIEW_USER_COUNTS = [
    {
        "ping_type": "*",
        "os": "*",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "*",
        "total_users": 44444,
    }
]

EXPECT = [
    {
        "app_build_id": "2020100100",
        "app_version": 84,
        "build_date": "2020-10-01 00:00:00",
        "channel": "*",
        "os": "*",
        "ping_type": "*",
        "total_users": 44444,
    }
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__view_user_counts_v1.yaml",
        VIEW_USER_COUNTS,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""
Determine BigQuery table storage.

To read more on the source table, please visit:
https://cloud.google.com/bigquery/docs/information-schema-table-storage
"""

from argparse import ArgumentParser

from google.cloud import bigquery

DEFAULT_PROJECTS = [
    "mozdata",
    "moz-fx-data-shared-prod",
    "moz-fx-data-marketing-prod",
]

parser = ArgumentParser(description=__doc__)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
# projects queries were run from that access table
parser.add_argument("--source_projects", nargs="+", default=DEFAULT_PROJECTS)
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="bigquery_table_storage_v1")


def create_query(source_project):
    """Create query for a source project.  1GB = POW(1024, 3) bytes."""
    return f"""
        SELECT
          DATE(creation_time) AS creation_date,
          project_id,
          table_schema AS dataset_id,
          table_name AS table_id,
          total_rows,
          total_partitions,
          total_logical_bytes,
          active_logical_bytes,
          long_term_logical_bytes,
          total_physical_bytes,
          active_physical_bytes,
          long_term_physical_bytes,
          time_travel_physical_bytes,
        FROM `{source_project}.region-us.INFORMATION_SCHEMA.TABLE_STORAGE`
        ORDER BY creation_date, project_id, dataset_id, table_id
    """


def main():
    """Run query for each source project."""
    args = parser.parse_args()

    destination_table = (
        f"{args.project}.{args.destination_dataset}.{args.destination_table}"
    )

    # remove old table in case of re-run
    client = bigquery.Client(args.project)
    client.delete_table(destination_table, not_found_ok=True)

    for project in args.source_projects:
        client = bigquery.Client(project)
        query = create_query(project)
        job_config = bigquery.QueryJobConfig(
            destination=destination_table, write_disposition="WRITE_APPEND"
        )
        client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



"""bigquery-etl CLI dryrun command."""

import fnmatch
import glob
import os
import re
import sys
from functools import partial
from multiprocessing.pool import Pool
from typing import List, Set, Tuple

import rich_click as click

from ..cli.utils import is_authenticated
from ..config import ConfigLoader
from ..dryrun import DryRun, get_credentials, get_id_token


@click.command(
    help="""Dry run SQL.
        Uses the dryrun Cloud Function by default which only has access to shared-prod.
        To dryrun queries accessing tables in another project use set
        `--use-cloud-function=false` and ensure that the command line has access to a
        GCP service account.

    Examples:
    ./bqetl dryrun sql/moz-fx-data-shared-prod/telemetry_derived/

    # Dry run SQL with tables that are not in shared prod
    ./bqetl dryrun --use-cloud-function=false sql/moz-fx-data-marketing-prod/
    """,
)
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(file_okay=True),
)
@click.option(
    "--use_cloud_function",
    "--use-cloud-function",
    help=(
        "Use the Cloud Function for dry running SQL, if set to `True`. "
        "The Cloud Function can only access tables in shared-prod. "
        "If set to `False`, use active GCP credentials for the dry run."
    ),
    type=bool,
    default=True,
)
@click.option(
    "--validate_schemas",
    "--validate-schemas",
    help="Require dry run schema to match destination table and file if present.",
    is_flag=True,
    default=False,
)
@click.option(
    "--respect-skip/--ignore-skip",
    help="Respect or ignore query skip configuration. Default is --respect-skip.",
    default=True,
)
@click.option(
    "--project",
    help="GCP project to perform dry run in when --use_cloud_function=False",
    default=ConfigLoader.get("default", "project", fallback="moz-fx-data-shared-prod"),
)
def dryrun(
    paths: List[str],
    use_cloud_function: bool,
    validate_schemas: bool,
    respect_skip: bool,
    project: str,
):
    """Perform a dry run."""
    file_names = (
        "query.sql",
        "view.sql",
        "part*.sql",
        "init.sql",
        "materialized_view.sql",
    )
    file_re = re.compile("|".join(map(fnmatch.translate, file_names)))

    sql_files: Set[str] = set()
    for path in paths:
        if os.path.isdir(path):
            sql_files |= {
                sql_file
                for pattern in file_names
                for sql_file in glob.glob(f"{path}/**/{pattern}", recursive=True)
            }
        elif os.path.isfile(path):
            if file_re.fullmatch(os.path.basename(path)):
                sql_files.add(path)
        else:
            click.echo(f"Invalid path {path}", err=True)
            sys.exit(1)
    if respect_skip:
        sql_files -= DryRun.skipped_files()

    if not sql_files:
        click.echo("Skipping dry run because no queries matched")
        sys.exit(0)

    if not use_cloud_function and not is_authenticated():
        click.echo(
            "Not authenticated to GCP. Run `gcloud auth login  --update-adc` to login."
        )
        sys.exit(1)

    credentials = get_credentials()
    id_token = get_id_token(credentials=credentials)

    sql_file_valid = partial(
        _sql_file_valid,
        use_cloud_function,
        respect_skip,
        validate_schemas,
        credentials=credentials,
        id_token=id_token,
    )

    with Pool(8) as p:
        result = p.map(sql_file_valid, sql_files, chunksize=1)

    failures = sorted([r[1] for r in result if not r[0]])
    if len(failures) > 0:
        click.echo(
            f"Failed to validate {len(failures)} queries (see above for error messages):",
            err=True,
        )
        click.echo("\n".join(failures), err=True)
        sys.exit(1)


def _sql_file_valid(
    use_cloud_function, respect_skip, validate_schemas, sqlfile, credentials, id_token
) -> Tuple[bool, str]:
    """Dry run the SQL file."""
    result = DryRun(
        sqlfile,
        use_cloud_function=use_cloud_function,
        credentials=credentials,
        respect_skip=respect_skip,
        id_token=id_token,
    )
    if validate_schemas:
        try:
            success = result.validate_schema()
        except Exception as e:  # validate_schema raises base exception
            click.echo(e, err=True)
            success = False
        return success, sqlfile

    return result.is_valid(), sqlfile



"""Generate serp events aggregates per app."""

import os
from enum import Enum
from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import render, write_sql

THIS_PATH = Path(os.path.dirname(__file__))
TABLE_NAME = "serp_events"


class Browsers(Enum):
    """Enumeration with browser names and equivalent dataset names."""

    firefox_desktop = "Firefox Desktop"


@click.command()
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="Google Cloud project ID",
    default="moz-fx-data-shared-prod",
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    """Generate per-app queries, views and metadata for serp events aggregates.

    The parent folders will be created if not existing and existing files will be overwritten.
    """
    env = Environment(loader=FileSystemLoader(str(THIS_PATH / "templates")))
    desktop_query_template = env.get_template("desktop_query.sql")
    metadata_template = "metadata.yaml"
    view_template = env.get_template("view.sql")
    output_dir = Path(output_dir) / target_project
    browser = Browsers["firefox_desktop"]
    query_sql = reformat(
        desktop_query_template.render(
            project_id=target_project,
            app_value=browser.value,
            app_name=browser.name,
        )
    )

    write_sql(
        output_dir=output_dir,
        full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}_v2",
        basename="query.sql",
        sql=query_sql,
        skip_existing=False,
    )

    write_sql(
        output_dir=output_dir,
        full_table_id=f"{target_project}.{browser.name}_derived.{TABLE_NAME}_v2",
        basename="metadata.yaml",
        sql=render(
            metadata_template,
            template_folder=THIS_PATH / "templates",
            app_value=browser.value,
            app_name=browser.name,
            format=False,
        ),
        skip_existing=False,
    )

    write_sql(
        output_dir=output_dir,
        full_table_id=f"{target_project}.{browser.name}.{TABLE_NAME}",
        basename="view.sql",
        sql=reformat(
            view_template.render(
                project_id=target_project,
                app_name=browser.name,
            )
        ),
        skip_existing=False,
    )



"""
Generate mobile search clients_daily query.

Create a combined CTE for metrics and baseline for Android and iOS Glean apps,
then print the query to a file in the output directory.
"""

from pathlib import Path
from typing import List

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import write_sql, render

# fmt: off
FIREFOX_ANDROID_TUPLES = [
    ("org_mozilla_fenix",           "Firefox Preview",  "beta"),  # noqa E241 E501
    ("org_mozilla_fenix_nightly",   "Firefox Preview",  "nightly"),  # noqa E241 E501
    ("org_mozilla_fennec_aurora",   "Fenix",            "nightly"),  # noqa E241 E501
    ("org_mozilla_firefox_beta",    "Fenix",            "beta"),  # noqa E241 E501
    ("org_mozilla_firefox",         "Fenix",            "release"),  # noqa E241 E501
]

DATASET = "search_derived"

TABLE_NAME = "mobile_search_clients_daily_v2"

FIREFOX_IOS_TUPLES = [
    ("org_mozilla_ios_firefox",     "Fennec", "release"),  # noqa E241 E501
    ("org_mozilla_ios_firefoxbeta", "Fennec", "beta"),  # noqa E241 E501
    ("org_mozilla_ios_fennec",      "Fennec", "nightly"),  # noqa E241 E501
]

FOCUS_ANDROID_TUPLES = [
    ("org_mozilla_focus",           "Focus Android Glean",    "release"),  # noqa E241 E501
    ("org_mozilla_focus_beta",      "Focus Android Glean",    "beta"),  # noqa E241 E501
    ("org_mozilla_focus_nightly",   "Focus Android Glean",    "nightly"),  # noqa E241 E501
]

KLAR_ANDROID_TUPLES = [
    ("org_mozilla_klar",            "Klar Android Glean",     "release"),  # noqa E241 E501
]

FOCUS_iOS_TUPLES = [
    ("org_mozilla_ios_focus",           "Focus iOS Glean",    "release"), # noqa E241 E501
]

KLAR_iOS_TUPLES = [
    ("org_mozilla_ios_klar",            "Klar iOS Glean",     "release"),  # noqa E241 E501
]
# fmt: on


def union_statements(statements: List[str]):
    """Join a list of strings together by UNION ALL."""
    return "\nUNION ALL\n".join(statements)


@click.command()
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="GCP project ID",
    default="moz-fx-data-shared-prod",
)
@use_cloud_function_option
def generate(output_dir, target_project, use_cloud_function):
    """Generate mobile search clients daily query and print to stdout."""

    base_dir = Path(__file__).parent

    output_dir = Path(output_dir) / target_project

    env = Environment(loader=FileSystemLoader(base_dir / "templates"))

    android_query_template = env.get_template("fenix.template.sql")
    ios_query_template = env.get_template("ios.template.sql")
    android_focus_template = env.get_template("android_focus.template.sql")
    android_klar_template = env.get_template("android_klar.template.sql")
    ios_focus_template = env.get_template("ios_focus.template.sql")
    ios_klar_template = env.get_template("ios_klar.template.sql")
    metadata_template = "mobile_search_clients_engines_sources_daily.metadata.yaml"

    firefox_android_queries = [
        android_query_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in FIREFOX_ANDROID_TUPLES
    ]

    firefox_ios_queries = [
        ios_query_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in FIREFOX_IOS_TUPLES
    ]

    focus_android_queries = [
        android_focus_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in FOCUS_ANDROID_TUPLES
    ]

    klar_android_queries = [
        android_klar_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in KLAR_ANDROID_TUPLES
    ]

    focus_ios_queries = [
        ios_focus_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in FOCUS_iOS_TUPLES
    ]

    klar_ios_queries = [
        ios_klar_template.render(
            namespace=namespace, app_name=app_name, channel=channel
        )
        for namespace, app_name, channel in KLAR_iOS_TUPLES
    ]

    queries = (
        firefox_android_queries
        + firefox_ios_queries
        + focus_android_queries
        + klar_android_queries
        + focus_ios_queries
        + klar_ios_queries
    )

    search_query_template = env.get_template("mobile_search_clients_daily.template.sql")


    fenix_combined_baseline = union_statements(
        [
            f"SELECT * FROM baseline_{namespace}"
            for namespace, _, _ in FIREFOX_ANDROID_TUPLES
        ]
    )
    
    ios_combined_baseline = union_statements(
        [f"SELECT * FROM baseline_{namespace}" for namespace, _, _ in FIREFOX_IOS_TUPLES]
    )
    android_focus_combined_baseline = union_statements(
        [
            f"SELECT * FROM baseline_{namespace}"
            for namespace, _, _ in FOCUS_ANDROID_TUPLES
        ]
    )
    android_klar_combined_baseline = union_statements(
        [
            f"SELECT * FROM baseline_{namespace}"
            for namespace, _, _ in KLAR_ANDROID_TUPLES
        ]
    )

    ios_focus_combined_baseline = union_statements(
        [f"SELECT * FROM baseline_{namespace}" for namespace, _, _ in FOCUS_iOS_TUPLES]
    )
    ios_klar_combined_baseline = union_statements(
        [f"SELECT * FROM baseline_{namespace}" for namespace, _, _ in KLAR_iOS_TUPLES]
    )

    search_query = search_query_template.render(
        baseline_by_namespace="\n".join(queries),
        fenix_baseline=fenix_combined_baseline,
        ios_metrics=ios_combined_baseline,
        android_focus_metrics=android_focus_combined_baseline,
        android_klar_metrics=android_klar_combined_baseline,
        ios_focus_metrics=ios_focus_combined_baseline,
        ios_klar_metrics=ios_klar_combined_baseline,
    )

    write_sql(
        output_dir=output_dir,
        full_table_id=f"{target_project}.{DATASET}.{TABLE_NAME}",
        basename="query.sql",
        sql=reformat(search_query),
        skip_existing=False,
    )

    write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{DATASET}.{TABLE_NAME}",
            basename="metadata.yaml",
            sql=render(
                metadata_template,
                template_folder=base_dir / "templates",
                format=False,
            ),
            skip_existing=False,
    )




"""Cinder API Addon Moderations Decisions - download decisions, clean and upload to BigQuery."""

import json
import os
import tempfile
from argparse import ArgumentParser

import requests
from google.cloud import bigquery

"""Get the bearer token for Cinder from the environment"""
cinder_bearer_token = os.environ.get("CINDER_TOKEN")


def post_response(url, headers, data):
    """POST response function."""
    response = requests.post(url, headers=headers, data=data)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def get_response(url, headers):
    """GET response function."""
    response = requests.get(url, headers=headers)
    if (response.status_code == 401) or (response.status_code == 400):
        print(f"***Error: {response.status_code}***")
        print(response.text)
    return response


def read_json(filename: str) -> dict:
    """Read JSON file."""
    with open(filename, "r") as f:
        data = json.loads(f.read())
    return data


def cinder_addon_decisions_download(date, bearer_token):
    """Download data from Cinder - bearer_token is called here."""
    url = "https://stage.cinder.nonprod.webservices.mozgcp.net/api/v1/decisions/"
    headers = {"accept": "application/json", "authorization": f"Bearer {bearer_token}"}
    print(url)
    response = get_response(url, headers)
    return response


def check_json(cinder_addon_decisions_response_text):
    """Script will return an empty dictionary for apps on days when there is no data. Check for that here."""
    with tempfile.NamedTemporaryFile() as tmp_json:
        with open(tmp_json.name, "w") as f_json:
            f_json.write(cinder_addon_decisions_response_text)
            try:
                query_export = read_json(f_json.name)
            except (
                ValueError
            ):  # ex. json.decoder.JSONDecodeError: Expecting value: line 1 column 1 (char 0)
                return None
    return query_export


def add_date_to_json(query_export_contents, date):
    """Add a date to the entries so we can partition table by this date."""
    fields_list = []
    for item in query_export_contents:
        field_dict = {
            "date": date,
            "decision_type": item["decision_type"],
            "user": item["user"],
            "created_at": item["created_at"],
            "entity_id": item["entity_id"],
            "entity": item["entity"],
            "uuid": item["uuid"],
            "entity_slug": item["entity_slug"],
            "job_id": item["job_id"],
            "job_assigned_at": item["job_assigned_at"],
            "typed_metadata": item["typed_metadata"],
            "applied_policies": item["applied_policies"],
        }
        fields_list.append(field_dict)
    return fields_list


def upload_to_bigquery(data, project, dataset, table_name, date):
    """Upload the data to bigquery."""
    date = date
    partition = f"{date}".replace("-", "")
    client = bigquery.Client(project)
    job_config = bigquery.LoadJobConfig(
        create_disposition="CREATE_IF_NEEDED",
        write_disposition="WRITE_TRUNCATE",
        time_partitioning=bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field="date",
        ),
    )
    destination = f"{project}.{dataset}.{table_name}${partition}"
    job = client.load_table_from_json(data, destination, job_config=job_config)
    job.result()


def main():
    """Input data, call functions, get stuff done."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="addon_moderations_derived")

    args = parser.parse_args()

    project = args.project
    dataset = args.dataset
    table_name = "cinder_decisions_raw_v1"

    date = args.date
    bearer_token = cinder_bearer_token

    cinder_data = []

    json_file = cinder_addon_decisions_download(date, bearer_token)
    """Data returns as a dictionary with a key called 'items' and the value being a list of data"""
    query_export = check_json(json_file.text)
    """Add date to each element in query_export for partitioning"""
    query_export_contents = query_export["items"]
    cinder_data = add_date_to_json(query_export_contents, date)
    """Pull out the list from query_export["items"] and put that data into the cinder_data list"""

    upload_to_bigquery(cinder_data, project, dataset, table_name, date)


if __name__ == "__main__":
    main()



"""Generate and upload JSON metadata files for public datasets on GCS."""

import json
import logging
import os
import re
from argparse import ArgumentParser
from itertools import groupby

import smart_open
from google.cloud import storage  # type: ignore

from bigquery_etl.config import ConfigLoader
from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.util import standard_args
from bigquery_etl.util.common import project_dirs

GCS_FILE_PATH_RE = re.compile(
    r"api/(?P<api_version>.+)/tables/(?P<dataset>.+)/(?P<table>.+)/(?P<version>.+)/"
    r"files/(?:(?P<date>.+)/)?(?P<filename>.+\.json(\.gz)?)"
)

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--project_id", "--project-id", default="mozilla-public-data", help="Target project"
)
parser.add_argument(
    "--target-bucket",
    "--target_bucket",
    default=ConfigLoader.get(
        "public_data", "bucket", fallback="mozilla-public-data-http"
    ),
    help="GCP bucket JSON data is exported to",
)
parser.add_argument(
    "--endpoint",
    default=ConfigLoader.get(
        "public_data", "endpoint", fallback="https://public-data.telemetry.mozilla.org/"
    ),
    help="The URL to access the HTTP endpoint",
)
parser.add_argument(
    "--api_version",
    "--api-version",
    default=ConfigLoader.get("public_data", "api_version", fallback="v1"),
    help="Endpoint API version",
)
standard_args.add_log_level(parser)


class GcsTableMetadata:
    """Metadata associated with table data stored on GCS."""

    def __init__(self, blobs, endpoint, target_dir):
        """Initialize container for metadata of a table published on GCS."""
        assert len(blobs) > 0
        self.blobs = blobs
        self.endpoint = endpoint
        self.files_path = self.blobs[0].name.split("files")[0] + "files"
        self.files_uri = endpoint + self.files_path

        (self.dataset, self.table, self.version) = dataset_table_version_from_gcs_blob(
            self.blobs[0]
        )
        self.metadata = Metadata.of_table(
            self.dataset, self.table, self.version, target_dir
        )

        self.last_updated_path = self.blobs[0].name.split("files")[0] + "last_updated"
        self.last_updated_uri = endpoint + self.last_updated_path

    def table_metadata_to_json(self):
        """Return a JSON object of the table metadata for GCS."""
        metadata_json = {}
        metadata_json["friendly_name"] = self.metadata.friendly_name
        metadata_json["description"] = self.metadata.description
        metadata_json["incremental"] = self.metadata.is_incremental()
        metadata_json["incremental_export"] = self.metadata.is_incremental_export()

        if self.metadata.review_bugs() is not None:
            review_links = list(
                map(
                    lambda bug: ConfigLoader.get(
                        "public_data",
                        "review_link",
                        fallback="https://bugzilla.mozilla.org/show_bug.cgi?id=",
                    )
                    + str(bug),
                    self.metadata.review_bugs(),
                )
            )
            metadata_json["review_links"] = review_links

        metadata_json["files_uri"] = self.files_uri
        metadata_json["last_updated"] = self.last_updated_uri

        return metadata_json

    def files_metadata_to_json(self):
        """Return a JSON object containing metadata of the files on GCS."""
        if self.metadata.is_incremental_export():
            metadata_json = {}

            for blob in self.blobs:
                match = GCS_FILE_PATH_RE.match(blob.name)
                date = match.group("date")

                if date is not None:
                    if date in metadata_json:
                        metadata_json[date].append(self.endpoint + blob.name)
                    else:
                        metadata_json[date] = [self.endpoint + blob.name]

            return metadata_json
        else:
            return [self.endpoint + blob.name for blob in self.blobs]


def dataset_table_version_from_gcs_blob(gcs_blob):
    """Extract the dataset, table and version from the provided GCS blob path."""
    match = GCS_FILE_PATH_RE.match(gcs_blob.name)

    if match is not None:
        return (match.group("dataset"), match.group("table"), match.group("version"))
    else:
        return None


def get_public_gcs_table_metadata(
    storage_client, bucket, api_version, endpoint, target_dir
):
    """Return a list of metadata of public tables and their locations on GCS."""
    prefix = f"api/{api_version}"

    blobs = storage_client.list_blobs(bucket, prefix=prefix)

    return [
        GcsTableMetadata(list(blobs), endpoint, target_dir)
        for table, blobs in groupby(blobs, dataset_table_version_from_gcs_blob)
        if table is not None
        and table[0] in os.listdir(target_dir)
        and f"{table[1]}_{table[2]}" in os.listdir(os.path.join(target_dir, table[0]))
    ]


def publish_all_datasets_metadata(table_metadata, output_file):
    """Write metadata about all available public datasets to GCS."""
    metadata_json = {}

    for metadata in table_metadata:
        if metadata.dataset not in metadata_json:
            metadata_json[metadata.dataset] = {}

        dataset = metadata_json[metadata.dataset]

        if metadata.table not in dataset:
            dataset[metadata.table] = {}

        table = dataset[metadata.table]

        if metadata.version not in table:
            table[metadata.version] = {}

        table[metadata.version] = metadata.table_metadata_to_json()

    logging.info(f"Write metadata to {output_file}")

    with smart_open.open(output_file, "w") as fout:
        fout.write(json.dumps(metadata_json, indent=4))


def publish_table_metadata(storage_client, table_metadata, bucket):
    """Write metadata for each public table to GCS."""
    for metadata in table_metadata:
        output_file = f"gs://{bucket}/{metadata.files_path}"

        logging.info(f"Write metadata to {output_file}")
        with smart_open.open(output_file, "w") as fout:
            fout.write(json.dumps(metadata.files_metadata_to_json(), indent=4))

        set_content_type(
            storage_client, bucket, f"{metadata.files_path}", "application/json"
        )


def set_content_type(storage_client, bucket, file, content_type):
    """Set the file content type."""
    blob = storage_client.get_bucket(bucket).get_blob(file)
    blob.content_type = content_type
    blob.patch()


def main():
    """Generate and upload GCS metadata."""
    args = parser.parse_args()
    storage_client = storage.Client(args.project_id)

    # set log level
    try:
        logging.basicConfig(level=args.log_level, format="%(levelname)s %(message)s")
    except ValueError as e:
        parser.error(f"argument --log-level: {e}")

    projects = project_dirs()
    all_metadata = []

    for target in projects:
        if os.path.isdir(target):
            gcs_table_metadata = get_public_gcs_table_metadata(
                storage_client,
                args.target_bucket,
                args.api_version,
                args.endpoint,
                target,
            )
            all_metadata += gcs_table_metadata
            publish_table_metadata(
                storage_client, gcs_table_metadata, args.target_bucket
            )
        else:
            print(
                f"Invalid target: {target}, target must be a directory with"
                "structure <project>/<dataset>/<table>/metadata.yaml."
            )

    output_file = f"gs://{args.target_bucket}/all-datasets.json"
    publish_all_datasets_metadata(all_metadata, output_file)
    set_content_type(
        storage_client,
        args.target_bucket,
        "all-datasets.json",
        "application/json",
    )


if __name__ == "__main__":
    main()



from pathlib import Path
from unittest.mock import patch

from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.publish_metadata import publish_metadata

TEST_DIR = Path(__file__).parent.parent


class TestPublishMetadata(object):
    @patch("google.cloud.bigquery.Client")
    @patch("google.cloud.bigquery.Table")
    def test_publish_metadata(self, mock_bigquery_table, mock_bigquery_client):
        mock_bigquery_client().get_table.return_value = mock_bigquery_table()

        metadata_file = (
            TEST_DIR
            / "data"
            / "test_sql"
            / "moz-fx-data-test-project"
            / "test"
            / "incremental_query_v1"
            / "metadata.yaml"
        )
        metadata = Metadata.from_file(metadata_file)

        project = "test_project"
        dataset = "test_dataset"
        table = "test_table"
        publish_metadata(mock_bigquery_client(), project, dataset, table, metadata)

        assert mock_bigquery_client().update_table.call_count == 1
        assert (
            mock_bigquery_client().update_table.call_args[0][0].friendly_name
            == "Test table for an incremental query"
        )
        assert (
            mock_bigquery_client().update_table.call_args[0][0].description
            == "Test table for an incremental query"
        )
        assert mock_bigquery_client().update_table.call_args[0][0].labels == {
            "dag": "bqetl_events",
            "deprecated": "true",
            "incremental": "",
            "incremental_export": "",
            "monitoring": "true",
            "owner1": "test",
            "public_json": "",
            "schedule": "daily",
        }
        assert (
            mock_bigquery_client()
            .get_table("moz-fx-data-test-project.test.incremental_query_v1")
            .friendly_name
            == "Test table for an incremental query"
        )
        assert (
            mock_bigquery_table().friendly_name == "Test table for an incremental query"
        )
        assert (
            mock_bigquery_table().description == "Test table for an incremental query"
        )
        assert mock_bigquery_table().labels == {
            "dag": "bqetl_events",
            "deprecated": "true",
            "incremental": "",
            "incremental_export": "",
            "monitoring": "true",
            "owner1": "test",
            "public_json": "",
            "schedule": "daily",
        }



"""Run a query with a series of @submission_date values."""

import os.path
import subprocess
import sys
from argparse import ArgumentParser
from datetime import datetime, timedelta
from functools import partial
from multiprocessing import Pool


def fromisoformat(string):
    """Construct a date from the output of date.isoformat()."""
    return datetime.strptime(string, "%Y-%m-%d").date()


def path(string):
    """Path to a file that must exist."""
    if not os.path.exists(string):
        raise ValueError()
    return string


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--destination_table",
    required=True,
    help="The destination table flag for `bq query` to which"
    " '$%%Y%%m%%d' will be added",
)
parser.add_argument(
    "--start",
    type=fromisoformat,
    help="The first date for which the query will be submitted,"
    " defaults to today in utc",
)
parser.add_argument(
    "--end",
    type=fromisoformat,
    help="The last date for which the query will be submitted,"
    " defaults to today in utc",
)
parser.add_argument(
    "--max_procs",
    "-P",
    type=int,
    default=1,
    help="Maximum number of queries which may be submitted at once",
)
parser.add_argument(
    "query_options",
    nargs="*",
    help="Arguments to pass directly to `bq query`",
)
parser.add_argument(
    "query_file", type=path, help="Path to the query file for `bq query`"
)


def _date_range(start=None, end=None, step=None):
    """Generate dates from START to END, in steps of STEP days."""
    if step is None:
        if end is not None and start is not None and end < start:
            step = -1
        else:
            step = 1
    if isinstance(step, int):
        step = timedelta(days=step)
    if end is None:
        end = datetime.utcnow().date()
    if start is None:
        start = end - step
    for offset in range((end - start) // step):
        yield start + (offset * step)


def _query(day, destination_table, query_options, query_file):
    with open(query_file) as query_stream:
        proc = subprocess.Popen(
            [
                "bq",
                "query",
                "--replace",
                "--dataset_id=telemetry",
                "--destination_table=" + destination_table + day.strftime("$%Y%m%d"),
                "--parameter=submission_date:DATE:" + day.isoformat(),
                "--max_rows=0",
                "--nouse_legacy_sql",
                "--quiet",
            ]
            + query_options,
            stdin=query_stream,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        prefix = day.isoformat() + ": "
        sys.stderr.write(prefix + "processing\n")
        for data, dest in zip(proc.communicate(), (sys.stdout, sys.stderr)):
            lines = data.split(b"\n")
            if len(lines[-1]) == 0:
                lines = lines[:-1]
            for line in lines:
                dest.write(prefix)
                dest.write(line.decode())
                dest.write("\n")
            dest.flush()
    return proc.returncode


def main():
    """Run a query with a series of @submission_date values."""
    args, unknown_args = parser.parse_known_args()
    # Unknown args should be passed as query_options.
    args.query_options.extend(unknown_args)
    target = partial(
        _query,
        destination_table=args.destination_table,
        query_options=args.query_options,
        query_file=args.query_file,
    )
    if args.max_procs == 1:
        for day in _date_range(args.start, args.end):
            target(day)
    else:
        Pool(processes=args.max_procs).map(
            target, _date_range(args.start, args.end), chunksize=1
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Import exchange rates from Oanda."""

import datetime
import json
from argparse import ArgumentParser
from dataclasses import dataclass
from decimal import Decimal

import requests
from google.cloud import bigquery


@dataclass(frozen=True)
class Quote:
    """Exchange rate quote."""

    date: datetime.date
    base_currency: str
    quote_currency: str
    price_type: str
    price: Decimal


parser = ArgumentParser(__doc__)
parser.add_argument("--start-date", required=True)
parser.add_argument("--end-date", required=True)
parser.add_argument("--table", required=True)
parser.add_argument(
    "--write-disposition",
    choices=[
        bigquery.WriteDisposition.WRITE_EMPTY,
        bigquery.WriteDisposition.WRITE_TRUNCATE,
        bigquery.WriteDisposition.WRITE_APPEND,
    ],
)
parser.add_argument(
    "--base-currencies",
    "--base-currency",
    nargs="+",
    required=True,
)
parser.add_argument("--quote-currency", default="USD")
parser.add_argument("--price", default="mid", choices=["bid", "ask", "mid"])
args = parser.parse_args()

quotes = set()
for base_currency in args.base_currencies:
    response = requests.get(
        "https://fxds-hcc.oanda.com/api/data/update/",
        params={
            "source": "OANDA",
            "adjustment": "0",
            "base_currency": base_currency,
            "start_date": args.start_date,
            "end_date": args.end_date,
            "period": "daily",
            "price": args.price,
            "view": "graph",
            "quote_currency_0": args.quote_currency,
        },
    )
    try:
        response.raise_for_status()
    except Exception:
        print(response.request)
        raise
    try:
        quotes |= {
            Quote(
                date=datetime.datetime.utcfromtimestamp(millis / 1000).date(),
                quote_currency=currency_pair["quoteCurrency"],
                base_currency=currency_pair["baseCurrency"],
                price_type=currency_pair["type"],
                price=round(Decimal(price), 9),
            )
            for currency_pair in response.json()["widget"]
            for millis, price in currency_pair["data"]
        }
    except Exception:
        print("invalid response:")
        print(response.text)
        raise

if not quotes:
    print("No exchange rates found.")
else:
    json_rows = [
        {
            "date": quote.date.isoformat(),
            "quote_currency": quote.quote_currency,
            "base_currency": quote.base_currency,
            "price_type": quote.price_type,
            "price": str(quote.price),
        }
        for quote in sorted(quotes, key=lambda q: q.date)
    ]
    print(*map(json.dumps, json_rows), sep="\n")
    table = args.table
    if args.start_date == args.end_date and "$" not in table:
        table = f"{table}${args.start_date.replace('-', '')}"
    print(f"writing to {table}")
    write_disposition = args.write_disposition
    if write_disposition is None:
        if "$" in table:
            write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        else:
            write_disposition = bigquery.WriteDisposition.WRITE_EMPTY
    job_config = bigquery.LoadJobConfig(
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
        schema=[
            bigquery.SchemaField("date", "DATE", "REQUIRED"),
            bigquery.SchemaField("base_currency", "STRING", "REQUIRED"),
            bigquery.SchemaField("quote_currency", "STRING", "REQUIRED"),
            bigquery.SchemaField("price_type", "STRING", "REQUIRED"),
            bigquery.SchemaField("price", "NUMERIC", "REQUIRED"),
        ],
        time_partitioning=bigquery.TimePartitioning(
            bigquery.TimePartitioningType.DAY, "date"
        ),
        write_disposition=write_disposition,
    )
    client = bigquery.Client()
    job = client.load_table_from_json(json_rows, table, job_config=job_config)
    job.result()



"""Generate unnested events queries for Glean apps."""

from sql_generators.glean_usage.common import GleanTable, ping_has_metrics

TARGET_TABLE_ID = "events_unnested_v1"
PREFIX = "events_unnested"
DATASET_SKIP = {
    # VPN events are in main pings
    "mozilla_vpn",
    "mozilla_vpn_android",
}


class EventsUnnestedTable(GleanTable):
    """Represents generated events_unnested table."""

    def __init__(self):
        """Initialize events_unnested table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.prefix = PREFIX
        self.per_app_id_enabled = False
        self.cross_channel_template = "cross_channel_events_unnested.view.sql"

    def generate_per_app(
        self,
        project_id,
        app_info,
        output_dir=None,
        use_cloud_function=True,
        parallelism=8,
        id_token=None,
    ):
        """Generate the events_unnested table query per app_name."""
        target_dataset = app_info[0]["app_name"]
        if target_dataset not in DATASET_SKIP:
            self.custom_render_kwargs = {
                "has_metrics": ping_has_metrics(
                    app_info[0]["bq_dataset_family"], "events"
                )
            }
            super().generate_per_app(project_id, app_info, output_dir, id_token=id_token)



"""Testing data for query."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent

UUID = "df735f02-efe5-4b07-b212-583bb99ba241"
SUBMISSION_DATE = "2020-10-01"
APP_BUILD_ID = "2020100100"

LATEST_VERSION = [dict(channel="nightly", latest_version=84)]

CLIENTS_HISTOGRAM_AGGREGATES = [
    {
        "sample_id": 1,
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "histogram_aggregates": [
            {
                "metric": "network_tcp_connection",
                "metric_type": "timing_distribution",
                "key": "",
                "agg_type": "summed_histogram",
                "value": [
                    {"key": "0", "value": 0},
                    {"key": "1", "value": 0},
                    {"key": "2", "value": 1},
                ],
            }
        ],
    }
]

CLIENTS_DAILY_HISTOGRAM_AGGREGATES = [
    {
        "submission_date": SUBMISSION_DATE,
        "sample_id": 1,
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "histogram_aggregates": [
            {
                "metric": "network_tcp_connection",
                "metric_type": "timing_distribution",
                "key": "",
                "agg_type": "summed_histogram",
                "value": [{"key": "0", "value": 1}, {"key": "1", "value": 0}],
            }
        ],
    }
]

EXPECT = [
    {
        "sample_id": 1,
        "client_id": UUID,
        "ping_type": "metrics",
        "os": "Android",
        "app_version": 84,
        "app_build_id": APP_BUILD_ID,
        "channel": "nightly",
        "histogram_aggregates": [
            {
                "metric": "network_tcp_connection",
                "metric_type": "timing_distribution",
                "key": "",
                "agg_type": "summed_histogram",
                "value": [
                    {"key": "0", "value": 1},
                    {"key": "1", "value": 0},
                    {"key": "2", "value": 1},
                ],
            }
        ],
    }
]

prefix = "glam_etl"
tables = [
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__latest_versions_v1.yaml",
        LATEST_VERSION,
    ),
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__clients_histogram_aggregates_v1.yaml",
        CLIENTS_HISTOGRAM_AGGREGATES,
    ),
    (
        f"{prefix}.org_mozilla_fenix_glam_nightly__"
        "view_clients_daily_histogram_aggregates_v1.yaml",
        CLIENTS_DAILY_HISTOGRAM_AGGREGATES,
    ),
    ("expect.yaml", EXPECT),
]
for name, data in tables:
    with (ROOT / name).open("w") as fp:
        yaml.dump(data, fp)



"""Generate signatures for symbolicated crash pings via siggen."""

from pathlib import Path

import click
import yaml
from google.cloud import bigquery, bigquery_storage
from siggen.generator import SignatureGenerator

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
OUTPUT_SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


@click.command()
@click.option(
    "--submission-date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    required=True,
)
@click.option(
    "--source-table",
    default="moz-fx-data-shared-prod.telemetry_derived.crash_symbolicated_v1",
)
@click.option(
    "--destination-table",
    default="moz-fx-data-shared-prod.telemetry_derived.crash_signatures_v1",
)
def main(
    submission_date,
    source_table,
    destination_table,
):
    sig_generator = SignatureGenerator()
    bq = bigquery.Client()
    bq_read = bigquery_storage.BigQueryReadClient()
    source = bq.get_table(source_table)
    session = bq_read.create_read_session(
        parent=f"projects/{source.project}",
        read_session=bigquery_storage.ReadSession(
            table=source.path.lstrip("/"),
            data_format=bigquery_storage.DataFormat.AVRO,
            read_options=bigquery_storage.ReadSession.TableReadOptions(
                row_restriction=f"DATE(submission_timestamp) = '{submission_date:%F}'",
            ),
        ),
        max_stream_count=1,
    )
    reader = bq_read.read_rows(session.streams[0].name)
    json_rows = [
        {
            "submission_timestamp": crash["submission_timestamp"].isoformat(),
            "document_id": crash["document_id"],
            "signature": sig_generator.generate(crash).signature or None,
        }
        for crash in reader.rows(session)
    ]

    job = bq.load_table_from_json(
        json_rows=json_rows,
        destination=f"{destination_table}${submission_date:%Y%m%d}",
        job_config=bigquery.LoadJobConfig(
            ignore_unknown_values=False,
            schema=OUTPUT_SCHEMA,
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        ),
    )
    job.result()


if __name__ == "__main__":
    main()



"""Represents a collection of configured Airflow DAGs."""

from collections import defaultdict
from functools import partial
from itertools import groupby
from multiprocessing import get_context, set_start_method
from operator import attrgetter
from pathlib import Path

import yaml
from black import FileMode, format_file_contents

from bigquery_etl.query_scheduling.dag import Dag, InvalidDag, PublicDataJsonDag


class DagCollection:
    """Representation of all configured DAGs."""

    def __init__(self, dags):
        """Instantiate DAGs."""
        self.dags = dags
        self.dags_by_name = {dag.name: dag for dag in dags}

    @classmethod
    def from_dict(cls, d):
        """
        Parse DAG configurations from a dict and create new instances.

        Expected dict format:
        {
            "bqetl_dag_name1": {
                "schedule_interval": string,
                "default_args": {
                    "owner": string,
                    "start_date": "YYYY-MM-DD",
                    ...
                }
            },
            "bqetl_dag_name2": {
                "schedule_interval": string,
                "default_args": {
                    "owner": string,
                    "start_date": "YYYY-MM-DD",
                    ...
                }
            },
            ...
        }
        """
        if d is None:
            return cls([])

        dags = [Dag.from_dict({k: v}) for k, v in d.items()]
        return cls(dags)

    @classmethod
    def from_file(cls, config_file):
        """Instantiate DAGs based on the provided configuration file."""
        with open(config_file, "r") as yaml_stream:
            dags_config = yaml.safe_load(yaml_stream)
            return DagCollection.from_dict(dags_config)

    def dag_by_name(self, name):
        """Return the DAG with the provided name."""
        return self.dags_by_name.get(name)

    def task_for_table(self, project, dataset, table):
        """Return the task that schedules the query for the provided table."""
        for dag in self.dags:
            for task in dag.tasks:
                if (
                    project == task.project
                    and dataset == task.dataset
                    and table == f"{task.table}_{task.version}"
                    and not task.is_dq_check
                    and not task.is_bigeye_check
                ):
                    return task

        return None

    def fail_checks_task_for_table(self, project, dataset, table):
        """Return the task that schedules the checks for the provided table."""
        for dag in self.dags:
            for task in dag.tasks:
                if (
                    project == task.project
                    and dataset == task.dataset
                    and table == f"{task.table}_{task.version}"
                    and task.is_dq_check
                    and task.is_dq_check_fail
                ):
                    return task

        return None

    def fail_bigeye_checks_task_for_table(self, project, dataset, table):
        """Return the task that schedules the BigEye checks for the provided table."""
        for dag in self.dags:
            for task in dag.tasks:
                if (
                    project == task.project
                    and dataset == task.dataset
                    and table == f"{task.table}_{task.version}"
                    and task.is_bigeye_check
                ):
                    return task

        return None

    def with_tasks(self, tasks):
        """Assign tasks to their corresponding DAGs."""
        public_data_json_dag = None

        get_dag_name = attrgetter("dag_name")
        for dag_name, group in groupby(sorted(tasks, key=get_dag_name), get_dag_name):
            dag = self.dag_by_name(dag_name)
            if dag is None:
                raise InvalidDag(
                    f"DAG {dag_name} does not exist in dags.yaml "
                    f"but used in task definition {next(group).task_name}."
                )
            dag.add_tasks(list(group))

        public_json_tasks = [
            task
            for task in tasks
            if task.public_json and not (task.is_dq_check or task.is_bigeye_check)
        ]
        if public_json_tasks:
            for dag in self.dags:
                if dag.__class__ == PublicDataJsonDag:
                    public_data_json_dag = dag
            if public_data_json_dag:
                public_data_json_dag.add_export_tasks(public_json_tasks, self)

        return self

    def get_task_downstream_dependencies(self, task):
        """Return all direct downstream dependencies of the task."""
        # Cache the downstream dependencies for faster lookups.
        if not hasattr(self, "_downstream_dependencies"):
            downstream_dependencies = defaultdict(list)
            for dag in self.dags:
                for _task in dag.tasks:
                    _task.with_upstream_dependencies(self)
                    for upstream_dependency in (
                        _task.depends_on + _task.upstream_dependencies
                    ):
                        downstream_dependencies[upstream_dependency.task_key].append(
                            _task.to_ref(self)
                        )
            self._downstream_dependencies = downstream_dependencies

        return self._downstream_dependencies[task.task_key]

    def dag_to_airflow(self, output_dir, dag):
        """Generate the Airflow DAG representation for the provided DAG."""
        output_file = Path(output_dir) / (dag.name + ".py")

        try:
            formatted_dag = format_file_contents(
                dag.to_airflow_dag(), fast=False, mode=FileMode()
            )
            output_file.write_text(formatted_dag)
        except InvalidDag as e:
            print(e)

    def to_airflow_dags(self, output_dir, dag_to_generate=None):
        """Write DAG representation as Airflow dags to file."""
        # https://pythonspeed.com/articles/python-multiprocessing/
        # when running tests on CI that call this function, we need
        # to create a custom pool to prevent processes from getting stuck

        # Generate a single DAG:
        if dag_to_generate is not None:
            dag_to_generate.with_upstream_dependencies(self)
            dag_to_generate.with_downstream_dependencies(self)
            self.dag_to_airflow(output_dir, dag_to_generate)
            return

        # Generate all DAGs:
        try:
            set_start_method("spawn")
        except Exception:
            pass

        for dag in self.dags:
            dag.with_upstream_dependencies(self)
            dag.with_downstream_dependencies(self)

        to_airflow_dag = partial(self.dag_to_airflow, output_dir)
        with get_context("spawn").Pool(8) as p:
            p.map(to_airflow_dag, self.dags)



"""Suppression List downloaded from CampaignMonitor API and uploaded to BigQuery."""

from pathlib import Path
from typing import Any, Dict, List

import requests
import rich_click as click
from google.cloud import bigquery
from requests import HTTPError
from requests.auth import HTTPBasicAuth

from bigquery_etl.schema import Schema

DEFAULT_PROJECT_ID = Path(__file__).parent.parent.parent.name
DEFAULT_DATASET_ID = Path(__file__).parent.parent.name
DEFAULT_TABLE_NAME = Path(__file__).parent.name

BASE_URL = "https://api.createsend.com/api/v3.3/clients"
SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = Schema.from_schema_file(SCHEMA_FILE).to_bigquery_schema()


def get_api_response(url: str, api_key: str):
    """Get json of response if the API request was successful."""
    headers = {"Accept": "application/json"}
    auth = HTTPBasicAuth(api_key, "")

    response = requests.get(url=url, headers=headers, auth=auth)
    try:
        response.raise_for_status()
    except HTTPError as error:
        raise Exception(f"API request {url} failed: {response.text}") from error
    return response.json()


def store_data_in_bigquery(data, schema, destination_project, destination_table_id):
    """Upload data to Bigquery in a single, non partitioned table."""
    client = bigquery.Client(project=destination_project)

    job_config = bigquery.LoadJobConfig(
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
        schema=schema,
        source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
    )

    load_job = client.load_table_from_json(
        data, destination_table_id, location="US", job_config=job_config
    )
    load_job.result()
    click.echo(f"Loaded {len(data)} rows into {destination_table_id}.")


def download_suppression_list(
    base_url: str,
    client_id: str,
    api_key: str,
    page_size: int = 1000,
    order_field: str = "date",
    order_direction: str = "asc",
) -> List[Dict[str, Any]]:
    """Download suppression list from Campaign Monitor."""
    page = 1
    suppressions_campaign_monitor_v1 = []

    while True:
        suppression_list_url = (
            f"{base_url}/{client_id}/suppressionlist.json"
            + f"?page={page}&pagesize={page_size}"
            + f"&orderfield={order_field}&orderdirection={order_direction}"
        )
        json_response = get_api_response(url=suppression_list_url, api_key=api_key)
        for result in json_response["Results"]:
            suppressions_campaign_monitor_v1.append(
                {
                    "suppression_reason": result["SuppressionReason"],
                    "email": result["EmailAddress"],
                    "update_timestamp": result["Date"],
                    "state": result["State"],
                }
            )

        if json_response["PageNumber"] >= json_response["NumberOfPages"]:
            return suppressions_campaign_monitor_v1
        page += 1


@click.command
@click.option(
    "--api_key",
    required=True,
    help="Campaign Monitor API key to use for authentication.",
)
@click.option(
    "--bq_project_id",
    default=DEFAULT_PROJECT_ID,
    show_default=True,
    help="BigQuery project the data is written to.",
)
@click.option(
    "--bq_dataset_id",
    default=DEFAULT_DATASET_ID,
    show_default=True,
    help="BigQuery dataset the data is written to.",
)
@click.option(
    "--bq_table_name",
    default=DEFAULT_TABLE_NAME,
    show_default=True,
    help="Bigquery table the data is written to.",
)
@click.option(
    "--client_id",
    required=True,
    help="Client Id for Campaign Monitor.",
)
def campaign_monitor_suppression_list_import_to_bigquery(
    api_key: str,
    bq_project_id: str,
    bq_dataset_id: str,
    bq_table_name: str,
    client_id: str,
):
    """Import Suppression List from Campaign Monitor."""
    suppressions = download_suppression_list(
        base_url=BASE_URL, client_id=client_id, api_key=api_key
    )
    destination_table_id = f"{bq_project_id}.{bq_dataset_id}.{bq_table_name}"
    store_data_in_bigquery(
        data=suppressions,
        schema=SCHEMA,
        destination_project=bq_project_id,
        destination_table_id=destination_table_id,
    )


if __name__ == "__main__":
    campaign_monitor_suppression_list_import_to_bigquery()



"""This file contains custom filters for formatting data types in Jinja templates."""

from datetime import datetime, timedelta

from bigquery_etl import query_scheduling
from bigquery_etl.query_scheduling.utils import TIMEDELTA_RE


def format_schedule_interval(interval):
    """Format the input value to a Airflow schedule_interval value."""
    presets = ["once", "hourly", "daily", "weekly", "monthly", "yearly"]

    if interval in presets:
        return "'@" + interval + "'"

    return repr(interval)


def format_attr(d, attribute, formatter_name):
    """Apply a formatter to a dict attribute."""
    for name in dir(query_scheduling.formatters):
        if name != formatter_name:
            continue

        formatter_func = getattr(query_scheduling.formatters, name)

        if callable(formatter_func) is None:
            continue

        if attribute in d:
            d[attribute] = formatter_func(d[attribute])

    return d


def format_date(date_string):
    """Format a date string to a datetime object."""
    if date_string is None:
        return None

    return datetime.strptime(date_string, "%Y-%m-%d")


# based on https://stackoverflow.com/questions/4628122/how-to-construct-a
# -timedelta-object-from-a-simple-string
def format_timedelta(timedelta_string):
    """Format a timedelta object."""
    parts = TIMEDELTA_RE.match(timedelta_string)
    if not parts:
        return timedelta_string

    parts = parts.groupdict()
    is_negative = timedelta_string.startswith("-")
    time_params = {}
    for name, param in parts.items():
        if param:
            time_params[name] = int(param) * (-1 if is_negative else 1)

    return timedelta(**time_params)


def format_timedelta_macro(timedelta_string):
    """Format an Airflow timedelta macro."""
    timedelta = repr(format_timedelta(timedelta_string))
    return timedelta.replace("datetime", "macros")


def format_optional_string(val):
    """Format a value that is either None or a string."""
    if val is None:
        return "None"
    else:
        return "'" + val + "'"


def format_repr(val):
    """Return representation of a object."""
    return repr(val)


def format_as_tuple(val):
    """Return tuple representation of the provided value."""
    return tuple(val)



from bigquery_etl.query_scheduling.utils import schedule_interval_delta


class TestUtils:
    def test_schedule_interval_delta(self):
        assert schedule_interval_delta("0 1 * * *", "0 2 * * *") == "3600s"
        assert schedule_interval_delta("0 2 * * *", "0 0 * * *") == "-7200s"
        assert schedule_interval_delta("* * * * *", "* * * * *") == "0s"
        assert schedule_interval_delta("0 1 * * *", "0 1 * * *") == "0s"
        assert schedule_interval_delta("daily", "0 1 * * *") == "3600s"
        assert schedule_interval_delta("daily", "0 0 * * *") == "0s"



#!/usr/bin/env python3

"""
Script for exporting data from client association ping from Firefox Desktop telemetry linked to a specific Account UID.

See https://bugzilla.mozilla.org/show_bug.cgi?id=1895503.
"""

import json
from argparse import ArgumentParser

from google.cloud import bigquery


def _clean_dict(obj):
    """
    Recursively remove keys with value of `None` or empty dictionaries.

    We have a lot of empty columns in the table, this makes the output smaller.
    """
    if isinstance(obj, dict):
        cleaned_dict = {
            k: _clean_dict(v) for k, v in obj.items() if v is not None and v != "None"
        }
        # Remove keys where the value is an empty dictionary
        return {k: v for k, v in cleaned_dict.items() if v}
    elif isinstance(obj, list):
        cleaned_list = [_clean_dict(item) for item in obj]
        # Filter out dictionaries where both 'key' and 'value' are None or empty lists
        cleaned_list = [
            item
            for item in cleaned_list
            if not (
                isinstance(item, dict)
                and item.get("key") is None
                and item.get("value") is None
            )
        ]
    else:
        return str(obj)


def main():
    """Export data, print to STDOUT."""
    parser = ArgumentParser(
        description="Export client association ping data from Firefox Desktop telemetry linked to a specific Account UID."
    )
    parser.add_argument("uid", type=str, help="Account UID to export data for")

    args = parser.parse_args()
    uid = args.uid

    print("Extracting client_association data for: " + uid)

    client = bigquery.Client()
    client_association_query = """
        SELECT
            *
        FROM
            `moz-fx-data-shared-prod.firefox_desktop_stable.fx_accounts_v1`
        WHERE
            DATE(submission_timestamp) >= '2024-06-01'
            AND metrics.string.client_association_uid = @uid
    """
    job_config = bigquery.QueryJobConfig(
        query_parameters=[bigquery.ScalarQueryParameter("uid", "STRING", uid)]
    )
    query_job = client.query(client_association_query, job_config=job_config)
    results = query_job.result()

    rows_as_dicts = [_clean_dict(dict(row)) for row in results]
    json_string = json.dumps(rows_as_dicts)

    print(json_string)


if __name__ == "__main__":
    main()



"""Fetch Analytics Report for Choice Screen Engagement in the App store via App Store Connect."""

import gzip
import io
import logging
import os
import tempfile
import time
from argparse import ArgumentParser
from pathlib import Path

import jwt
import pandas as pd
import requests
from google.cloud import bigquery

from bigquery_etl.schema import Schema

#
# https://developer.apple.com/documentation/AppStoreConnectAPI/downloading-analytics-reports
CONNECT_ISSUER_ID = os.environ.get("CONNECT_ISSUER_ID")
CONNECT_KEY_ID = os.environ.get("CONNECT_KEY_ID")
CONNECT_KEY = os.environ.get("CONNECT_KEY")

HOST = "https://api.appstoreconnect.apple.com"
REPORT_TITLE = "Browser Choice Screen Engagement (iOS versions before 18.2)"
REPORT_DATA_DELIMITER = "\t"

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
SCHEMA = Schema.from_schema_file(SCHEMA_FILE).to_bigquery_schema()


def generate_jwt_token(issuer_id, key_id, private_key):
    """Generate jtw token to be used for API authentication."""
    payload = {
        "iss": issuer_id,
        "iat": int(time.time()),
        "exp": int(time.time()) + 20 * 60,  # Token valid for 20 minutes
        "aud": "appstoreconnect-v1",
    }
    algorithm = "ES256"

    return jwt.encode(
        payload,
        private_key.replace("\\n", "\n"),
        algorithm=algorithm,
        headers={"kid": key_id, "alg": algorithm, "typ": "JWT"},
    )


def download_gz_file(url, target_file_path):
    """Download and decompress a gz file."""
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        file_content = io.BytesIO()

        for chunk in response.iter_content(chunk_size=8192):
            if chunk:  # Filter out keep-alive new chunks
                file_content.write(chunk)

        # # Move the file content pointer to the beginning for reading
        file_content.seek(0)

        # # Decompress from in-memory buffer to file
        with gzip.GzipFile(fileobj=file_content, mode="rb") as f_in:
            with open(target_file_path, "wb") as f_out:
                f_out.write(f_in.read())


def api_call(url, jwt_token):
    """Handle making API requests to App Store Connect."""
    headers = {
        "Authorization": f"Bearer {jwt_token}",
        "Content-Type": "application/json",
    }

    response = requests.get(url, headers=headers)
    response.raise_for_status()

    return response.json()


def get_paginated_data(url, jwt_token):
    """Get data from all result pages."""
    data = []
    while True:
        result = api_call(url, jwt_token)
        data.extend(result["data"])
        if next := result["links"].get("next"):
            url = next
        else:
            return data


def fetch_app_reports(jwt_token, app_id):
    """Retrieve the report requests for the given app."""
    endpoint = f"apps/{app_id}/analyticsReportRequests"
    response_data = api_call(f"{HOST}/v1/{endpoint}", jwt_token)
    return response_data.get("data")


def fetch_report_instances(jwt_token, report_id):
    """Retrieve a list of report instances from the App Store."""
    endpoint = f"v1/analyticsReportInstances/{report_id}/segments"
    response_data = api_call(f"{HOST}/{endpoint}", jwt_token)
    return response_data.get("data")


def fetch_report_data(app_id, date, jwt_token, target_file_path):
    """Fetch and prints the report data."""
    # "ONGOING" provides current data and generates reports daily, weekly and monthly.
    # assumed there is only 1 ongoing report request at any given time
    ongoing_analytics_report_request = list(
        filter(
            lambda x: x["attributes"]["accessType"] == "ONGOING",
            fetch_app_reports(jwt_token, app_id),
        )
    )[0]

    analytics_reports = get_paginated_data(
        ongoing_analytics_report_request["relationships"]["reports"]["links"][
            "related"
        ],
        jwt_token,
    )

    # for now we just assume there is only one ongoing browser choice screen report
    browser_choice_screen_analytics_report = list(
        filter(
            lambda x: x["attributes"]["name"] == REPORT_TITLE,
            analytics_reports,
        )
    )[0]

    report_instances = api_call(
        browser_choice_screen_analytics_report["relationships"]["instances"]["links"][
            "related"
        ],
        jwt_token,
    )["data"]

    daily_report_instances = list(
        filter(lambda x: x["attributes"]["granularity"] == "DAILY", report_instances)
    )
    specific_date_report_instance = list(
        filter(
            lambda x: x["attributes"]["processingDate"] == date, daily_report_instances
        )
    )[0]

    analytics_report_segments = api_call(
        specific_date_report_instance["relationships"]["segments"]["links"]["related"],
        jwt_token,
    )["data"][0]

    checksum = analytics_report_segments["attributes"]["checksum"]

    response = requests.get(analytics_report_segments["attributes"]["url"])
    response.raise_for_status()

    download_gz_file(
        analytics_report_segments["attributes"]["url"],
        target_file_path=target_file_path,
    )

    return target_file_path, checksum


def upload_to_bigquery(
    local_file_path, project, dataset, table_name, date, partition_field
):
    """Upload the data to bigquery."""
    df = pd.read_csv(local_file_path, delimiter=REPORT_DATA_DELIMITER)
    df[partition_field] = date
    df.columns = [x.lower().replace(" ", "_") for x in df.columns]

    df["date"] = pd.to_datetime(df["date"], format="%Y-%m-%d")
    df[partition_field] = pd.to_datetime(df["logical_date"], format="%Y-%m-%d")

    job_config = bigquery.LoadJobConfig(
        create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
        write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
        schema=SCHEMA,
    )

    client = bigquery.Client(project)
    destination = f"{project}.{dataset}.{table_name}${date.replace('-', '')}"

    job = client.load_table_from_dataframe(df, destination, job_config=job_config)

    while client.get_job(job.job_id, location=job.location).state != "DONE":
        logging.info("Waiting for the bq load job to be done, job_id: %s." % job.job_id)
        time.sleep(5)

    return job.result(), destination


def main():
    """Input data, call functions, get stuff done."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", required=True)
    parser.add_argument(
        "--connect_app_id", required=True, help="App Store Connect App ID"
    )
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--dataset", default="firefox_ios_derived")
    parser.add_argument("--table", default="app_store_choice_screen_engagement_v1")
    parser.add_argument("--partition_field", default="logical_date")
    parser.add_argument(
        "--connect_issuer_id",
        default=CONNECT_ISSUER_ID,
        help="App Store Connect Issuer ID",
    )
    parser.add_argument(
        "--connect_key_id", default=CONNECT_KEY_ID, help="App Store Connect Key ID"
    )
    parser.add_argument(
        "--connect_private_key",
        default=CONNECT_KEY,
        help="Private key",
    )
    args = parser.parse_args()

    jwt_token = generate_jwt_token(
        args.connect_issuer_id, args.connect_key_id, args.connect_private_key
    )

    with tempfile.NamedTemporaryFile() as temp_file:
        report_file, _ = fetch_report_data(
            app_id=args.connect_app_id,
            date=args.date,
            jwt_token=jwt_token,
            target_file_path=temp_file.name,
        )

        logging.info("Report file downloaded: %s" % report_file)

        _, destination_table = upload_to_bigquery(
            temp_file.name,
            args.project,
            args.dataset,
            args.table,
            args.date,
            args.partition_field,
        )

        logging.info("BigQuery table has been created: %s" % destination_table)


if __name__ == "__main__":
    main()



"""Generate and run metrics_clients_last_seen queries for Glean apps."""

import os
from pathlib import Path

import yaml

from sql_generators.glean_usage.common import GleanTable

TARGET_TABLE_ID = "metrics_clients_last_seen_v1"
PREFIX = "metrics_clients_last_seen"


class MetricsClientsLastSeen(GleanTable):
    """Represents generated metrics_clients_last_seen table."""

    def __init__(self):
        """Initialize metrics_clients_last_seen table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.per_app_id_enabled = False
        self.cross_channel_template = None

        with open(
            Path(os.path.dirname(__file__)) / "templates" / "metrics_templating.yaml",
            "r",
        ) as f:
            metrics_config = yaml.safe_load(f) or {}
            self.custom_render_kwargs = {"metrics": metrics_config}



"""Metric-hub integration."""

from typing import Dict, List, Optional, Union

import attr
from metric_config_parser.config import ConfigCollection


@attr.s(auto_attribs=True, slots=True)
class MetricHub:
    """Metric-hub integration for generating SQL from referenced metrics."""

    _config_collection: Optional[ConfigCollection] = None

    @property
    def config_collection(self):
        """Config collection instance."""
        self._config_collection = (
            self._config_collection or ConfigCollection.from_github_repo()
        )
        return self._config_collection

    def calculate(
        self,
        metrics: List[str],
        platform: str,
        group_by: Union[List[str], Dict[str, str]] = [],
        where: Optional[str] = None,
        group_by_client_id: bool = True,
        group_by_submission_date: bool = True,
    ) -> str:
        """Generate SQL query for specified metrics."""
        return self.config_collection.get_metrics_sql(
            metrics=metrics,
            platform=platform,
            group_by=group_by,
            where=where,
            group_by_client_id=group_by_client_id,
            group_by_submission_date=group_by_submission_date,
        )

    def data_source(
        self,
        data_source: str,
        platform: str,
        where: Optional[str] = None,
    ) -> str:
        """Generate SQL query for specified data source."""
        return self.config_collection.get_data_source_sql(
            data_source=data_source, platform=platform, where=where
        )



from pathlib import Path

import yaml

ROOT_DIR = Path(__file__).parent.parent.parent


def test_dag_names():
    """Verify dag names.

    DAG names defined in this repository must start with `bqetl_`.
    CLI allows to create DAGs prefixed with `bqetl_` and `private_bqetl_`,
    the latter are reserved for private version of bigquery-etl repository.
    """
    with open(str(ROOT_DIR / "dags.yaml"), "r") as dags_file:
        dags_conf = yaml.safe_load(dags_file.read())
    for dag_name in dags_conf:
        assert dag_name.startswith("bqetl_")



#!/usr/bin/env python3

"""Run a query script to check if bigquery-etl is working properly."""

from argparse import ArgumentParser

from google.cloud import bigquery

parser = ArgumentParser(description=__doc__)
parser.add_argument("--run_id", required=True)
parser.add_argument("--task_instance", required=True)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="bigquery_etl_python_run_check_v1")


def main():
    args = parser.parse_args()

    destination_table = (
        f"{args.project}.{args.destination_dataset}.{args.destination_table}"
    )

    client = bigquery.Client(args.project)
    job_config = bigquery.QueryJobConfig(
        destination=destination_table, write_disposition="WRITE_TRUNCATE"
    )

    query = f"""
      SELECT
        CURRENT_TIMESTAMP() AS run_timestamp,
        '{args.task_instance}' AS task_instance,
        '{args.run_id}' AS run_id
    """

    client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



"""Represents an Airflow DAG."""

from typing import List, Optional, Set

import attr
import cattrs
from jinja2 import Environment, PackageLoader

from bigquery_etl.config import ConfigLoader
from bigquery_etl.query_scheduling import formatters
from bigquery_etl.query_scheduling.task import Task, TaskRef
from bigquery_etl.query_scheduling.utils import (
    is_date_string,
    is_email_or_github_identity,
    is_schedule_interval,
    is_valid_dag_name,
    schedule_interval_delta,
    validate_timedelta_string,
)

AIRFLOW_DAG_TEMPLATE = "airflow_dag.j2"
PUBLIC_DATA_JSON_DAG_TEMPLATE = "public_data_json_airflow_dag.j2"
PUBLIC_DATA_JSON_DAG = "bqetl_public_data_json"

CONFIDENTIAL_TAG = "triage/confidential"


class DagParseException(Exception):
    """Raised when DAG config is invalid."""

    def __init__(self, message):
        """Throw DagParseException."""
        message = f"""
        {message}

        Expected yaml format:
        name:
            schedule_interval: string,
            default_args:
                owner: string
                start_date: 'YYYY-MM-DD'
                ...
        """

        super(DagParseException, self).__init__(message)


class InvalidDag(Exception):
    """Raised when the resulting DAG is invalid."""

    pass


@attr.s(auto_attribs=True)
class DagDefaultArgs:
    """
    Representation of Airflow DAG default_args.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    owner: str = attr.ib()
    start_date: str = attr.ib()
    end_date: Optional[str] = attr.ib(None)
    email: List[str] = attr.ib([])
    depends_on_past: bool = attr.ib(False)
    retry_delay: str = attr.ib("30m")
    email_on_failure: bool = attr.ib(True)
    email_on_retry: bool = attr.ib(True)
    retries: int = attr.ib(2)

    @owner.validator
    def validate_owner(self, attribute, value):
        """Check that owner is a valid email address."""
        if not is_email_or_github_identity(value):
            raise ValueError(
                f"Invalid email or github identity for DAG owner: {value}."
            )

    @email.validator
    def validate_email(self, attribute, value):
        """Check that provided email addresses are valid."""
        if not all(map(lambda e: is_email_or_github_identity(e), value)):
            raise ValueError(f"Invalid email or github identity in DAG email: {value}.")

    @retry_delay.validator
    def validate_retry_delay(self, attribute, value):
        """Check that retry_delay is in a valid timedelta format."""
        validate_timedelta_string(value)

    @end_date.validator
    @start_date.validator
    def validate_date(self, attribute, value):
        """Check that start_date and end_date has YYYY-MM-DD format."""
        if value is not None and not is_date_string(value):
            raise ValueError(
                f"Invalid date definition for {attribute}: {value}."
                "Dates should be specified as YYYY-MM-DD."
            )

    def to_dict(self):
        """Return class as a dict."""
        return self.__dict__


@attr.s(auto_attribs=True)
class Dag:
    """
    Representation of a DAG configuration.

    Uses attrs to simplify the class definition and provide validation.
    Docs: https://www.attrs.org
    """

    name: str = attr.ib()
    schedule_interval: str = attr.ib()
    default_args: DagDefaultArgs
    tasks: List[Task] = attr.ib([])
    description: str = attr.ib("")
    repo: str = attr.ib("bigquery-etl")
    tags: List[str] = attr.ib([])

    @name.validator
    def validate_dag_name(self, attribute, value):
        """Validate the DAG name."""
        if not is_valid_dag_name(value):
            raise ValueError(
                f"Invalid DAG name {value}. Name must start with 'bqetl_' "
                f"or 'private_bqetl_'."
            )

    @tasks.validator
    def validate_tasks(self, attribute, value):
        """Validate tasks."""
        task_names = list(map(lambda t: t.task_name, value))
        duplicate_task_names = set(
            [task_name for task_name in task_names if task_names.count(task_name) > 1]
        )

        if len(duplicate_task_names) > 0:
            raise ValueError(
                f"Duplicate task names encountered: {duplicate_task_names}."
            )

    @schedule_interval.validator
    def validate_schedule_interval(self, attribute, value):
        """Validate the schedule_interval format."""
        if not is_schedule_interval(value):
            raise ValueError(f"Invalid schedule_interval {value}.")

    def add_tasks(self, tasks):
        """Add tasks to be scheduled as part of the DAG."""
        self.tasks = self.tasks.copy() + tasks
        self.validate_tasks(None, self.tasks)

    def with_upstream_dependencies(self, dag_collection):
        """Perform a dry_run to get upstream dependencies."""
        for task in self.tasks:
            task.with_upstream_dependencies(dag_collection)

    def with_downstream_dependencies(self, dag_collection):
        """Get downstream tasks by looking up upstream dependencies in DAG collection."""
        for task in self.tasks:
            task.with_downstream_dependencies(dag_collection)

    def to_dict(self):
        """Return class as a dict."""
        d = self.__dict__
        name = d["name"]
        del d["name"]
        del d["tasks"]
        d["default_args"] = self.default_args.to_dict()

        return {name: d}

    @property
    def task_groups(self) -> Set[str]:
        """
        Return list of task groups in this DAG.

        Task groups are specified as part of the task configurations.
        """
        return {task.task_group for task in self.tasks if task.task_group is not None}

    @classmethod
    def from_dict(cls: type, d: dict):
        """
        Parse the DAG configuration from a dict and create a new Dag instance.

        Expected dict format:
        {
            "name": {
                "schedule_interval": string,
                "default_args": dict
            }
        }
        """
        if len(d.keys()) != 1:
            raise DagParseException(f"Invalid DAG configuration format in {d}")

        converter = cattrs.BaseConverter()
        try:
            name = list(d.keys())[0]
            tags: set[str] = set(d[name].get("tags", []))
            if not any(tag.startswith("repo/") for tag in tags):
                tags.add("repo/" + d[name].get("repo", "bigquery-etl"))

            if name.startswith("private_") and CONFIDENTIAL_TAG not in tags:
                tags.add(
                    CONFIDENTIAL_TAG,
                )

            d[name]["tags"] = sorted(tags)

            if name == PUBLIC_DATA_JSON_DAG:
                return converter.structure({"name": name, **d[name]}, PublicDataJsonDag)
            else:
                return converter.structure({"name": name, **d[name]}, cls)
        except (TypeError, AttributeError) as e:
            raise DagParseException(f"Invalid DAG configuration format in {d}: {e}")

    def _jinja_env(self):
        """Prepare and load custom formatters into the jinja environment."""
        env = Environment(
            loader=PackageLoader("bigquery_etl", "query_scheduling/templates"),
            extensions=["jinja2.ext.do"],
        )

        # load custom formatters into Jinja env
        for name in dir(formatters):
            func = getattr(formatters, name)
            if not callable(func):
                continue

            env.filters[name] = func

        return env

    def to_airflow_dag(self):
        """Convert the DAG to its Airflow representation and return the python code."""
        if len(self.tasks) == 0:
            raise InvalidDag(
                f"DAG {self.name} has no tasks - cannot convert it to a valid .py DAG "
                f"file. Does it appear under `scheduling` in any metadata.yaml files?"
            )

        env = self._jinja_env()
        dag_template = env.get_template(AIRFLOW_DAG_TEMPLATE)
        args = self.__dict__
        args["task_groups"] = self.task_groups
        args["bigeye_warehouse_id"] = ConfigLoader.get(
            "monitoring", "bigeye_warehouse_id", fallback=1939
        )
        args["bigeye_conn_id"] = ConfigLoader.get(
            "monitoring", "bigeye_conn_id", fallback="bigeye_connection"
        )

        return dag_template.render(args)


class PublicDataJsonDag(Dag):
    """Special DAG with tasks exporting public json data to GCS."""

    def to_airflow_dag(self):
        """Convert the DAG to its Airflow representation and return the python code."""
        env = self._jinja_env()
        dag_template = env.get_template(PUBLIC_DATA_JSON_DAG_TEMPLATE)
        args = self.__dict__

        return dag_template.render(args)

    def _create_export_task(self, task, dag_collection):
        if not task.public_json:
            raise ValueError(f"Task {task.task_name} not marked as public JSON.")

        converter = cattrs.BaseConverter()
        task_dict = converter.unstructure(task)

        del task_dict["dataset"]
        del task_dict["table"]
        del task_dict["version"]
        del task_dict["project"]

        export_task = converter.structure(task_dict, Task)
        export_task.dag_name = self.name
        export_task.task_name = f"export_public_data_json_{export_task.task_name}"

        task_schedule_interval = dag_collection.dag_by_name(
            task.dag_name
        ).schedule_interval

        execution_delta = schedule_interval_delta(
            task_schedule_interval, self.schedule_interval
        )

        if execution_delta == "0s":
            execution_delta = None

        export_task.dependencies = [
            TaskRef(
                dag_name=task.dag_name,
                task_id=task.task_name,
                execution_delta=execution_delta,
            )
        ]

        return export_task

    def add_export_tasks(self, tasks, dag_collection):
        """Add new tasks for exporting data of the original queries to GCS."""
        self.add_tasks(
            [self._create_export_task(task, dag_collection) for task in tasks]
        )



#!/usr/bin/env python3

"""
Determine BigQuery usage.

To read more on the source table, please visit:
https://cloud.google.com/bigquery/docs/information-schema-jobs
"""

from argparse import ArgumentParser

from google.cloud import bigquery

DEFAULT_PROJECTS = [
    "mozdata",
    "moz-fx-data-shared-prod",
    "moz-fx-data-marketing-prod",
]

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-shared-prod")
# projects queries were run from that access table
parser.add_argument("--source_projects", nargs="+", default=DEFAULT_PROJECTS)
parser.add_argument("--destination_dataset", default="monitoring_derived")
parser.add_argument("--destination_table", default="bigquery_usage_v1")

# total_bytes_billed a minimum of 10MB per table total_bytes_processed


def create_query(date, source_project):
    """Create query for a source project."""
    return f"""
        SELECT
         "{source_project}" AS source_project,
          DATE('{date}') AS creation_date,
          job_id,
          job_type,
          reservation_id,
          cache_hit,
          state,
          referenced_tables.project_id AS reference_project_id,
          dataset_id AS reference_dataset_id,
          table_id AS reference_table_id,
          destination_table.project_id AS  destination_project_id,
          destination_table.dataset_id AS  destination_dataset_id,
          destination_table.table_id AS  destination_table_id,
          user_email,
          REGEXP_EXTRACT(query, r"Username: (.*?),") AS username,
          REGEXP_EXTRACT(query, r"Query ID: (\\w+), ") AS query_id,
          end_time-start_time as task_duration,
          ROUND(total_bytes_processed / 1024 / 1024 / 1024 / 1024, 4)
          AS total_terabytes_processed,
          error_result.location AS error_location,
          error_result.reason AS error_reason,
          error_result.message AS error_message,
        FROM
          `{source_project}.region-us.INFORMATION_SCHEMA.JOBS_BY_PROJECT`
        LEFT JOIN
          UNNEST(referenced_tables) AS referenced_tables
        WHERE
          DATE(creation_time) = '{date}'
    """


def main():
    """Run query for each source project."""
    args = parser.parse_args()

    partition = args.date.replace("-", "")
    destination_table = f"{args.project}.{args.destination_dataset}.{args.destination_table}${partition}"

    # remove old partition in case of re-run
    client = bigquery.Client(args.project)
    client.delete_table(destination_table, not_found_ok=True)

    for project in args.source_projects:
        client = bigquery.Client(project)
        query = create_query(args.date, project)
        job_config = bigquery.QueryJobConfig(
            destination=destination_table, write_disposition="WRITE_APPEND"
        )
        client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Import VAT rates from Google Sheets."""

import google.auth
from google.cloud import bigquery

# Use job config with external table definition
external_config = bigquery.ExternalConfig("GOOGLE_SHEETS")
external_config.source_uris = [
    "https://docs.google.com/spreadsheets/d/1-eUbaYIuppfwoCyawKEfNVHRx_HJfn_VBOc7c2YYS48"
]
external_config.schema = [
    bigquery.SchemaField("country_code", "STRING"),
    bigquery.SchemaField("country", "STRING"),
    bigquery.SchemaField("vat", "STRING"),
    bigquery.SchemaField("effective_date", "DATE"),
]
external_config.options.skip_leading_rows = 2
job_config = bigquery.QueryJobConfig(
    table_definitions={"vat_rates_sheet": external_config}
)

# Use credentials that include a google drive scope
credentials, project = google.auth.default(
    scopes=[
        "https://www.googleapis.com/auth/drive",
        "https://www.googleapis.com/auth/bigquery",
    ]
)
client = bigquery.Client(credentials=credentials, project=project)

query = client.query(
    """CREATE OR REPLACE TABLE
  `moz-fx-data-shared-prod`.mozilla_vpn_derived.vat_rates_v1
AS
SELECT
  * REPLACE (
    COALESCE(
      CAST(REGEXP_EXTRACT(vat, r"(.*)%") AS NUMERIC) / 100,
      CAST(vat AS NUMERIC)
    ) AS vat
  )
FROM
  vat_rates_sheet""",
    job_config=job_config,
)
query.result()



import os

import pytest
from click.testing import CliRunner
from jinja2.exceptions import TemplateNotFound

from bigquery_etl.dependency import show as dependency_show


class TestDependency:
    @pytest.fixture
    def runner(self):
        return CliRunner()

    def test_format_invalid_path(self, runner):
        result = runner.invoke(dependency_show, ["not-existing-path.sql"])
        assert result.exit_code == 1
        assert isinstance(result.exception, TemplateNotFound)

    def test_format(self, runner):
        with runner.isolated_filesystem():
            with open("foo.sql", "w") as f:
                f.write("SELECT 1 FROM test")

            result = runner.invoke(dependency_show, ["foo.sql"])
            assert "foo.sql: test\n" == result.output
            assert result.exit_code == 0

            os.mkdir("test")
            with open("test/foo.sql", "w") as f:
                f.write("SELECT 1 FROM test_foo")
            with open("test/bar.sql", "w") as f:
                f.write("SELECT 1 FROM test_bar")

            result = runner.invoke(dependency_show, ["test"])
            assert "test/bar.sql: test_bar\ntest/foo.sql: test_foo\n" == result.output
            assert result.exit_code == 0
